###########################################################################################
# Implementation of MACE models and other models based E(3)-Equivariant MPNNs
# Authors: Ilyes Batatia, Gregor Simm
# This program is distributed under the MIT License (see MIT.md)
###########################################################################################

from typing import Any, Callable, Dict, List, Optional, Type, Union

import numpy as np
import torch
from macer.externals.e3nn_legacy import o3
from macer.externals.e3nn_legacy.util.jit import compile_mode

from macer.externals.mace_bundled.modules.embeddings import GenericJointEmbedding
from macer.externals.mace_bundled.modules.radial import ZBLBasis
from macer.externals.mace_bundled.tools.scatter import scatter_mean, scatter_sum
from macer.externals.mace_bundled.tools.torch_tools import get_change_of_basis, spherical_to_cartesian

from .blocks import (
    AtomicEnergiesBlock,
    EquivariantProductBasisBlock,
    InteractionBlock,
    LinearDipolePolarReadoutBlock,
    LinearDipoleReadoutBlock,
    LinearNodeEmbeddingBlock,
    LinearReadoutBlock,
    NonLinearDipolePolarReadoutBlock,
    NonLinearDipoleReadoutBlock,
    NonLinearReadoutBlock,
    RadialEmbeddingBlock,
    ScaleShiftBlock,
)
from .utils import (
    compute_dielectric_gradients,
    compute_fixed_charge_dipole,
    compute_fixed_charge_dipole_polar,
    get_atomic_virials_stresses,
    get_edge_vectors_and_lengths,
    get_outputs,
    get_symmetric_displacement,
    prepare_graph,
)


@compile_mode("script")
class MACE(torch.nn.Module):
    def __init__(
        self,
        r_max: float,
        num_bessel: int,
        num_polynomial_cutoff: int,
        max_ell: int,
        interaction_cls: Type[InteractionBlock],
        interaction_cls_first: Type[InteractionBlock],
        num_interactions: int,
        num_elements: int,
        hidden_irreps: o3.Irreps,
        MLP_irreps: o3.Irreps,
        atomic_energies: np.ndarray,
        avg_num_neighbors: float,
        atomic_numbers: List[int],
        correlation: Union[int, List[int]],
        gate: Optional[Callable],
        pair_repulsion: bool = False,
        apply_cutoff: bool = True,
        use_reduced_cg: bool = True,
        use_so3: bool = False,
        use_agnostic_product: bool = False,
        use_last_readout_only: bool = False,
        use_embedding_readout: bool = False,
        distance_transform: str = "None",
        edge_irreps: Optional[o3.Irreps] = None,
        radial_MLP: Optional[List[int]] = None,
        radial_type: Optional[str] = "bessel",
        heads: Optional[List[str]] = None,
        cueq_config: Optional[Dict[str, Any]] = None,
        embedding_specs: Optional[Dict[str, Any]] = None,
        oeq_config: Optional[Dict[str, Any]] = None,
        lammps_mliap: Optional[bool] = False,
        readout_cls: Optional[Type[NonLinearReadoutBlock]] = NonLinearReadoutBlock,
    ):
        super().__init__()
        self.register_buffer(
            "atomic_numbers", torch.tensor(atomic_numbers, dtype=torch.int64)
        )
        self.register_buffer(
            "r_max", torch.tensor(r_max, dtype=torch.get_default_dtype())
        )
        self.register_buffer(
            "num_interactions", torch.tensor(num_interactions, dtype=torch.int64)
        )
        if heads is None:
            heads = ["Default"]
        self.heads = heads
        if isinstance(correlation, int):
            correlation = [correlation] * num_interactions
        self.lammps_mliap = lammps_mliap
        self.apply_cutoff = apply_cutoff
        self.edge_irreps = edge_irreps
        self.use_reduced_cg = use_reduced_cg
        self.use_agnostic_product = use_agnostic_product
        self.use_so3 = use_so3
        self.use_last_readout_only = use_last_readout_only

        # Embedding
        node_attr_irreps = o3.Irreps([(num_elements, (0, 1))])
        node_feats_irreps = o3.Irreps([(hidden_irreps.count(o3.Irrep(0, 1)), (0, 1))])
        self.node_embedding = LinearNodeEmbeddingBlock(
            irreps_in=node_attr_irreps,
            irreps_out=node_feats_irreps,
            cueq_config=cueq_config,
        )
        embedding_size = node_feats_irreps.count(o3.Irrep(0, 1))
        if embedding_specs is not None:
            self.embedding_specs = embedding_specs
            self.joint_embedding = GenericJointEmbedding(
                base_dim=embedding_size,
                embedding_specs=embedding_specs,
                out_dim=embedding_size,
            )
            if use_embedding_readout:
                self.embedding_readout = LinearReadoutBlock(
                    node_feats_irreps,
                    o3.Irreps(f"{len(heads)}x0e"),
                    cueq_config,
                    oeq_config,
                )

        self.radial_embedding = RadialEmbeddingBlock(
            r_max=r_max,
            num_bessel=num_bessel,
            num_polynomial_cutoff=num_polynomial_cutoff,
            radial_type=radial_type,
            distance_transform=distance_transform,
            apply_cutoff=apply_cutoff,
        )
        edge_feats_irreps = o3.Irreps(f"{self.radial_embedding.out_dim}x0e")
        if pair_repulsion:
            self.pair_repulsion_fn = ZBLBasis(p=num_polynomial_cutoff)
            self.pair_repulsion = True

        if not use_so3:
            sh_irreps = o3.Irreps.spherical_harmonics(max_ell)
        else:
            sh_irreps = o3.Irreps.spherical_harmonics(max_ell, p=1)
        num_features = hidden_irreps.count(o3.Irrep(0, 1))

        # interaction_irreps = (sh_irreps * num_features).sort()[0].simplify()
        def generate_irreps(l):
            str_irrep = "+".join([f"1x{i}e+1x{i}o" for i in range(l + 1)])
            return o3.Irreps(str_irrep)

        sh_irreps_inter = sh_irreps
        if hidden_irreps.count(o3.Irrep(0, -1)) > 0:
            sh_irreps_inter = generate_irreps(max_ell)
        interaction_irreps = (sh_irreps_inter * num_features).sort()[0].simplify()
        interaction_irreps_first = (sh_irreps * num_features).sort()[0].simplify()

        self.spherical_harmonics = o3.SphericalHarmonics(
            sh_irreps, normalize=True, normalization="component"
        )
        if radial_MLP is None:
            radial_MLP = [64, 64, 64]
        # Interactions and readout
        self.atomic_energies_fn = AtomicEnergiesBlock(atomic_energies)

        inter = interaction_cls_first(
            node_attrs_irreps=node_attr_irreps,
            node_feats_irreps=node_feats_irreps,
            edge_attrs_irreps=sh_irreps,
            edge_feats_irreps=edge_feats_irreps,
            target_irreps=interaction_irreps_first,
            hidden_irreps=hidden_irreps,
            avg_num_neighbors=avg_num_neighbors,
            radial_MLP=radial_MLP,
            cueq_config=cueq_config,
            oeq_config=oeq_config,
        )
        self.interactions = torch.nn.ModuleList([inter])

        # Use the appropriate self connection at the first layer for proper E0
        use_sc_first = False
        if "Residual" in str(interaction_cls_first):
            use_sc_first = True

        node_feats_irreps_out = inter.target_irreps
        prod = EquivariantProductBasisBlock(
            node_feats_irreps=node_feats_irreps_out,
            target_irreps=hidden_irreps,
            correlation=correlation[0],
            num_elements=num_elements,
            use_sc=use_sc_first,
            cueq_config=cueq_config,
            oeq_config=oeq_config,
            use_reduced_cg=use_reduced_cg,
            use_agnostic_product=use_agnostic_product,
        )
        self.products = torch.nn.ModuleList([prod])

        self.readouts = torch.nn.ModuleList()
        if not use_last_readout_only:
            self.readouts.append(
                LinearReadoutBlock(
                    hidden_irreps,
                    o3.Irreps(f"{len(heads)}x0e"),
                    cueq_config,
                    oeq_config,
                )
            )

        for i in range(num_interactions - 1):
            if i == num_interactions - 2:
                hidden_irreps_out = str(
                    hidden_irreps[0]
                )  # Select only scalars for last layer
            else:
                hidden_irreps_out = hidden_irreps
            inter = interaction_cls(
                node_attrs_irreps=node_attr_irreps,
                node_feats_irreps=hidden_irreps,
                edge_attrs_irreps=sh_irreps,
                edge_feats_irreps=edge_feats_irreps,
                target_irreps=interaction_irreps,
                hidden_irreps=hidden_irreps_out,
                avg_num_neighbors=avg_num_neighbors,
                edge_irreps=edge_irreps,
                radial_MLP=radial_MLP,
                cueq_config=cueq_config,
                oeq_config=oeq_config,
            )
            self.interactions.append(inter)
            prod = EquivariantProductBasisBlock(
                node_feats_irreps=interaction_irreps,
                target_irreps=hidden_irreps_out,
                correlation=correlation[i + 1],
                num_elements=num_elements,
                use_sc=True,
                cueq_config=cueq_config,
                oeq_config=oeq_config,
                use_reduced_cg=use_reduced_cg,
                use_agnostic_product=use_agnostic_product,
            )
            self.products.append(prod)
            if i == num_interactions - 2:
                self.readouts.append(
                    readout_cls(
                        hidden_irreps_out,
                        (len(heads) * MLP_irreps).simplify(),
                        gate,
                        o3.Irreps(f"{len(heads)}x0e"),
                        len(heads),
                        cueq_config,
                        oeq_config,
                    )
                )
            elif not use_last_readout_only:
                self.readouts.append(
                    LinearReadoutBlock(
                        hidden_irreps,
                        o3.Irreps(f"{len(heads)}x0e"),
                        cueq_config,
                        oeq_config,
                    )
                )

    def forward(
        self,
        data: Dict[str, torch.Tensor],
        training: bool = False,
        compute_force: bool = True,
        compute_virials: bool = False,
        compute_stress: bool = False,
        compute_displacement: bool = False,
        compute_hessian: bool = False,
        compute_edge_forces: bool = False,
        compute_atomic_stresses: bool = False,
        lammps_mliap: bool = False,
    ) -> Dict[str, Optional[torch.Tensor]]:
        # Setup
        ctx = prepare_graph(
            data,
            compute_virials=compute_virials,
            compute_stress=compute_stress,
            compute_displacement=compute_displacement,
            lammps_mliap=lammps_mliap,
        )
        is_lammps = ctx.is_lammps
        num_atoms_arange = ctx.num_atoms_arange.to(torch.int64)
        num_graphs = ctx.num_graphs
        displacement = ctx.displacement
        positions = ctx.positions
        vectors = ctx.vectors
        lengths = ctx.lengths
        cell = ctx.cell
        node_heads = ctx.node_heads.to(torch.int64)
        interaction_kwargs = ctx.interaction_kwargs
        lammps_natoms = interaction_kwargs.lammps_natoms
        lammps_class = interaction_kwargs.lammps_class

        # Atomic energies
        node_e0 = self.atomic_energies_fn(data["node_attrs"])[
            num_atoms_arange, node_heads
        ]
        e0 = scatter_sum(
            src=node_e0, index=data["batch"], dim=0, dim_size=num_graphs
        ).to(
            vectors.dtype
        )  # [n_graphs, n_heads]
        # Embeddings
        node_feats = self.node_embedding(data["node_attrs"])
        edge_attrs = self.spherical_harmonics(vectors)
        edge_feats, cutoff = self.radial_embedding(
            lengths, data["node_attrs"], data["edge_index"], self.atomic_numbers
        )
        if hasattr(self, "pair_repulsion"):
            pair_node_energy = self.pair_repulsion_fn(
                lengths, data["node_attrs"], data["edge_index"], self.atomic_numbers
            )
            if is_lammps:
                pair_node_energy = pair_node_energy[: lammps_natoms[0]]
            pair_energy = scatter_sum(
                src=pair_node_energy, index=data["batch"], dim=-1, dim_size=num_graphs
            )  # [n_graphs,]
        else:
            pair_node_energy = torch.zeros_like(node_e0)
            pair_energy = torch.zeros_like(e0)

        if hasattr(self, "joint_embedding"):
            embedding_features: Dict[str, torch.Tensor] = {}
            for name, _ in self.embedding_specs.items():
                embedding_features[name] = data[name]
            node_feats += self.joint_embedding(
                data["batch"],
                embedding_features,
            )
            if hasattr(self, "embedding_readout"):
                embedding_node_energy = self.embedding_readout(
                    node_feats, node_heads
                ).squeeze(-1)
                embedding_energy = scatter_sum(
                    src=embedding_node_energy,
                    index=data["batch"],
                    dim=0,
                    dim_size=num_graphs,
                )
                e0 += embedding_energy

        # Interactions
        energies = [e0, pair_energy]
        node_energies_list = [node_e0, pair_node_energy]
        node_feats_concat: List[torch.Tensor] = []

        for i, (interaction, product) in enumerate(
            zip(self.interactions, self.products)
        ):
            node_attrs_slice = data["node_attrs"]
            if is_lammps and i > 0:
                node_attrs_slice = node_attrs_slice[: lammps_natoms[0]]
            node_feats, sc = interaction(
                node_attrs=node_attrs_slice,
                node_feats=node_feats,
                edge_attrs=edge_attrs,
                edge_feats=edge_feats,
                edge_index=data["edge_index"],
                cutoff=cutoff,
                first_layer=(i == 0),
                lammps_class=lammps_class,
                lammps_natoms=lammps_natoms,
            )
            if is_lammps and i == 0:
                node_attrs_slice = node_attrs_slice[: lammps_natoms[0]]
            node_feats = product(
                node_feats=node_feats, sc=sc, node_attrs=node_attrs_slice
            )
            node_feats_concat.append(node_feats)

        for i, readout in enumerate(self.readouts):
            feat_idx = -1 if len(self.readouts) == 1 else i
            node_es = readout(node_feats_concat[feat_idx], node_heads)[
                num_atoms_arange, node_heads
            ]
            energy = scatter_sum(node_es, data["batch"], dim=0, dim_size=num_graphs)
            energies.append(energy)
            node_energies_list.append(node_es)

        contributions = torch.stack(energies, dim=-1)
        total_energy = torch.sum(contributions, dim=-1)
        node_energy = torch.sum(torch.stack(node_energies_list, dim=-1), dim=-1)
        node_feats_out = torch.cat(node_feats_concat, dim=-1)

        forces, virials, stress, hessian, edge_forces = get_outputs(
            energy=total_energy,
            positions=positions,
            displacement=displacement,
            vectors=vectors,
            cell=cell,
            training=training,
            compute_force=compute_force,
            compute_virials=compute_virials,
            compute_stress=compute_stress,
            compute_hessian=compute_hessian,
            compute_edge_forces=compute_edge_forces,
        )

        atomic_virials: Optional[torch.Tensor] = None
        atomic_stresses: Optional[torch.Tensor] = None
        if compute_atomic_stresses and edge_forces is not None:
            atomic_virials, atomic_stresses = get_atomic_virials_stresses(
                edge_forces=edge_forces,
                edge_index=data["edge_index"],
                vectors=vectors,
                num_atoms=positions.shape[0],
                batch=data["batch"],
                cell=cell,
            )
        return {
            "energy": total_energy,
            "node_energy": node_energy,
            "contributions": contributions,
            "forces": forces,
            "edge_forces": edge_forces,
            "virials": virials,
            "stress": stress,
            "atomic_virials": atomic_virials,
            "atomic_stresses": atomic_stresses,
            "displacement": displacement,
            "hessian": hessian,
            "node_feats": node_feats_out,
        }


@compile_mode("script")
class ScaleShiftMACE(MACE):
    def __init__(
        self,
        atomic_inter_scale: float,
        atomic_inter_shift: float,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.scale_shift = ScaleShiftBlock(
            scale=atomic_inter_scale, shift=atomic_inter_shift
        )

    def forward(
        self,
        data: Dict[str, torch.Tensor],
        training: bool = False,
        compute_force: bool = True,
        compute_virials: bool = False,
        compute_stress: bool = False,
        compute_displacement: bool = False,
        compute_hessian: bool = False,
        compute_edge_forces: bool = False,
        compute_atomic_stresses: bool = False,
        lammps_mliap: bool = False,
    ) -> Dict[str, Optional[torch.Tensor]]:
        # Setup
        ctx = prepare_graph(
            data,
            compute_virials=compute_virials,
            compute_stress=compute_stress,
            compute_displacement=compute_displacement,
            lammps_mliap=lammps_mliap,
        )

        is_lammps = ctx.is_lammps
        num_atoms_arange = ctx.num_atoms_arange.to(torch.int64)
        num_graphs = ctx.num_graphs
        displacement = ctx.displacement
        positions = ctx.positions
        vectors = ctx.vectors
        lengths = ctx.lengths
        cell = ctx.cell
        node_heads = ctx.node_heads.to(torch.int64)
        interaction_kwargs = ctx.interaction_kwargs
        lammps_natoms = interaction_kwargs.lammps_natoms
        lammps_class = interaction_kwargs.lammps_class

        # Atomic energies
        node_e0 = self.atomic_energies_fn(data["node_attrs"])[
            num_atoms_arange, node_heads
        ]
        e0 = scatter_sum(
            src=node_e0, index=data["batch"], dim=0, dim_size=num_graphs
        ).to(
            vectors.dtype
        )  # [n_graphs, num_heads]

        # Embeddings
        node_feats = self.node_embedding(data["node_attrs"])
        edge_attrs = self.spherical_harmonics(vectors)
        edge_feats, cutoff = self.radial_embedding(
            lengths, data["node_attrs"], data["edge_index"], self.atomic_numbers
        )

        if hasattr(self, "pair_repulsion"):
            pair_node_energy = self.pair_repulsion_fn(
                lengths, data["node_attrs"], data["edge_index"], self.atomic_numbers
            )
            if is_lammps:
                pair_node_energy = pair_node_energy[: lammps_natoms[0]]
        else:
            pair_node_energy = torch.zeros_like(node_e0)

        # Embeddings of additional features
        if hasattr(self, "joint_embedding"):
            embedding_features: Dict[str, torch.Tensor] = {}
            for name, _ in self.embedding_specs.items():
                embedding_features[name] = data[name]
            node_feats += self.joint_embedding(
                data["batch"],
                embedding_features,
            )
            if hasattr(self, "embedding_readout"):
                embedding_node_energy = self.embedding_readout(
                    node_feats, node_heads
                ).squeeze(-1)
                embedding_energy = scatter_sum(
                    src=embedding_node_energy,
                    index=data["batch"],
                    dim=0,
                    dim_size=num_graphs,
                )
                e0 += embedding_energy

        # Interactions
        node_es_list = [pair_node_energy]
        node_feats_list: List[torch.Tensor] = []

        for i, (interaction, product) in enumerate(
            zip(self.interactions, self.products)
        ):
            node_attrs_slice = data["node_attrs"]
            if is_lammps and i > 0:
                node_attrs_slice = node_attrs_slice[: lammps_natoms[0]]
            node_feats, sc = interaction(
                node_attrs=node_attrs_slice,
                node_feats=node_feats,
                edge_attrs=edge_attrs,
                edge_feats=edge_feats,
                edge_index=data["edge_index"],
                cutoff=cutoff,
                first_layer=(i == 0),
                lammps_class=lammps_class,
                lammps_natoms=lammps_natoms,
            )
            if is_lammps and i == 0:
                node_attrs_slice = node_attrs_slice[: lammps_natoms[0]]
            node_feats = product(
                node_feats=node_feats, sc=sc, node_attrs=node_attrs_slice
            )
            node_feats_list.append(node_feats)

        for i, readout in enumerate(self.readouts):
            feat_idx = -1 if len(self.readouts) == 1 else i
            node_es_list.append(
                readout(node_feats_list[feat_idx], node_heads)[
                    num_atoms_arange, node_heads
                ]
            )

        node_feats_out = torch.cat(node_feats_list, dim=-1)
        node_inter_es = torch.sum(torch.stack(node_es_list, dim=0), dim=0)
        node_inter_es = self.scale_shift(node_inter_es, node_heads)
        inter_e = scatter_sum(node_inter_es, data["batch"], dim=-1, dim_size=num_graphs)

        total_energy = e0 + inter_e
        node_energy = node_e0.clone().double() + node_inter_es.clone().double()

        forces, virials, stress, hessian, edge_forces = get_outputs(
            energy=inter_e,
            positions=positions,
            displacement=displacement,
            vectors=vectors,
            cell=cell,
            training=training,
            compute_force=compute_force,
            compute_virials=compute_virials,
            compute_stress=compute_stress,
            compute_hessian=compute_hessian,
            compute_edge_forces=compute_edge_forces or compute_atomic_stresses,
        )

        atomic_virials: Optional[torch.Tensor] = None
        atomic_stresses: Optional[torch.Tensor] = None
        if compute_atomic_stresses and edge_forces is not None:
            atomic_virials, atomic_stresses = get_atomic_virials_stresses(
                edge_forces=edge_forces,
                edge_index=data["edge_index"],
                vectors=vectors,
                num_atoms=positions.shape[0],
                batch=data["batch"],
                cell=cell,
            )
        return {
            "energy": total_energy,
            "node_energy": node_energy,
            "interaction_energy": inter_e,
            "forces": forces,
            "edge_forces": edge_forces,
            "virials": virials,
            "stress": stress,
            "atomic_virials": atomic_virials,
            "atomic_stresses": atomic_stresses,
            "hessian": hessian,
            "displacement": displacement,
            "node_feats": node_feats_out,
        }


@compile_mode("script")
class AtomicDipolesMACE(torch.nn.Module):
    def __init__(
        self,
        r_max: float,
        num_bessel: int,
        num_polynomial_cutoff: int,
        max_ell: int,
        interaction_cls: Type[InteractionBlock],
        interaction_cls_first: Type[InteractionBlock],
        num_interactions: int,
        num_elements: int,
        hidden_irreps: o3.Irreps,
        MLP_irreps: o3.Irreps,
        avg_num_neighbors: float,
        atomic_numbers: List[int],
        correlation: int,
        gate: Optional[Callable],
        atomic_energies: Optional[
            None
        ],  # Just here to make it compatible with energy models, MUST be None
        apply_cutoff: bool = True,  # pylint: disable=unused-argument
        use_reduced_cg: bool = True,  # pylint: disable=unused-argument
        use_so3: bool = False,  # pylint: disable=unused-argument
        distance_transform: str = "None",  # pylint: disable=unused-argument
        radial_type: Optional[str] = "bessel",
        radial_MLP: Optional[List[int]] = None,
        cueq_config: Optional[Dict[str, Any]] = None,  # pylint: disable=unused-argument
        oeq_config: Optional[Dict[str, Any]] = None,  # pylint: disable=unused-argument
        edge_irreps: Optional[o3.Irreps] = None,  # pylint: disable=unused-argument
    ):
        super().__init__()
        self.register_buffer(
            "atomic_numbers", torch.tensor(atomic_numbers, dtype=torch.int64)
        )
        self.register_buffer("r_max", torch.tensor(r_max, dtype=torch.float64))
        self.register_buffer(
            "num_interactions", torch.tensor(num_interactions, dtype=torch.int64)
        )
        assert atomic_energies is None

        # Embedding
        node_attr_irreps = o3.Irreps([(num_elements, (0, 1))])
        node_feats_irreps = o3.Irreps([(hidden_irreps.count(o3.Irrep(0, 1)), (0, 1))])
        self.node_embedding = LinearNodeEmbeddingBlock(
            irreps_in=node_attr_irreps, irreps_out=node_feats_irreps
        )
        self.radial_embedding = RadialEmbeddingBlock(
            r_max=r_max,
            num_bessel=num_bessel,
            num_polynomial_cutoff=num_polynomial_cutoff,
            radial_type=radial_type,
        )
        edge_feats_irreps = o3.Irreps(f"{self.radial_embedding.out_dim}x0e")

        sh_irreps = o3.Irreps.spherical_harmonics(max_ell)
        num_features = hidden_irreps.count(o3.Irrep(0, 1))
        interaction_irreps = (sh_irreps * num_features).sort()[0].simplify()
        self.spherical_harmonics = o3.SphericalHarmonics(
            sh_irreps, normalize=True, normalization="component"
        )
        if radial_MLP is None:
            radial_MLP = [64, 64, 64]

        # Interactions and readouts
        inter = interaction_cls_first(
            node_attrs_irreps=node_attr_irreps,
            node_feats_irreps=node_feats_irreps,
            edge_attrs_irreps=sh_irreps,
            edge_feats_irreps=edge_feats_irreps,
            target_irreps=interaction_irreps,
            hidden_irreps=hidden_irreps,
            avg_num_neighbors=avg_num_neighbors,
            radial_MLP=radial_MLP,
        )
        self.interactions = torch.nn.ModuleList([inter])

        # Use the appropriate self connection at the first layer
        use_sc_first = False
        if "Residual" in str(interaction_cls_first):
            use_sc_first = True

        node_feats_irreps_out = inter.target_irreps
        prod = EquivariantProductBasisBlock(
            node_feats_irreps=node_feats_irreps_out,
            target_irreps=hidden_irreps,
            correlation=correlation,
            num_elements=num_elements,
            use_sc=use_sc_first,
        )
        self.products = torch.nn.ModuleList([prod])

        self.readouts = torch.nn.ModuleList()
        self.readouts.append(LinearDipoleReadoutBlock(hidden_irreps, dipole_only=True))

        for i in range(num_interactions - 1):
            if i == num_interactions - 2:
                assert (
                    len(hidden_irreps) > 1
                ), "To predict dipoles use at least l=1 hidden_irreps"
                hidden_irreps_out = str(
                    hidden_irreps[1]
                )  # Select only l=1 vectors for last layer
            else:
                hidden_irreps_out = hidden_irreps
            inter = interaction_cls(
                node_attrs_irreps=node_attr_irreps,
                node_feats_irreps=hidden_irreps,
                edge_attrs_irreps=sh_irreps,
                edge_feats_irreps=edge_feats_irreps,
                target_irreps=interaction_irreps,
                hidden_irreps=hidden_irreps_out,
                avg_num_neighbors=avg_num_neighbors,
                radial_MLP=radial_MLP,
            )
            self.interactions.append(inter)
            prod = EquivariantProductBasisBlock(
                node_feats_irreps=interaction_irreps,
                target_irreps=hidden_irreps_out,
                correlation=correlation,
                num_elements=num_elements,
                use_sc=True,
            )
            self.products.append(prod)
            if i == num_interactions - 2:
                self.readouts.append(
                    NonLinearDipoleReadoutBlock(
                        hidden_irreps_out, MLP_irreps, gate, dipole_only=True
                    )
                )
            else:
                self.readouts.append(
                    LinearDipoleReadoutBlock(hidden_irreps, dipole_only=True)
                )

    def forward(
        self,
        data: Dict[str, torch.Tensor],
        training: bool = False,  # pylint: disable=W0613
        compute_force: bool = False,
        compute_virials: bool = False,
        compute_stress: bool = False,
        compute_displacement: bool = False,
        compute_edge_forces: bool = False,  # pylint: disable=W0613
        compute_atomic_stresses: bool = False,  # pylint: disable=W0613
    ) -> Dict[str, Optional[torch.Tensor]]:
        assert compute_force is False
        assert compute_virials is False
        assert compute_stress is False
        assert compute_displacement is False
        # Setup
        data["node_attrs"].requires_grad_(True)
        data["positions"].requires_grad_(True)
        num_graphs = data["ptr"].numel() - 1

        # Embeddings
        node_feats = self.node_embedding(data["node_attrs"])
        vectors, lengths = get_edge_vectors_and_lengths(
            positions=data["positions"],
            edge_index=data["edge_index"],
            shifts=data["shifts"],
        )
        edge_attrs = self.spherical_harmonics(vectors)
        edge_feats, cutoff = self.radial_embedding(
            lengths, data["node_attrs"], data["edge_index"], self.atomic_numbers
        )

        # Interactions
        dipoles = []
        for interaction, product, readout in zip(
            self.interactions, self.products, self.readouts
        ):
            node_feats, sc = interaction(
                node_attrs=data["node_attrs"],
                node_feats=node_feats,
                edge_attrs=edge_attrs,
                edge_feats=edge_feats,
                edge_index=data["edge_index"],
                cutoff=cutoff,
            )
            node_feats = product(
                node_feats=node_feats,
                sc=sc,
                node_attrs=data["node_attrs"],
            )
            node_dipoles = readout(node_feats).squeeze(-1)  # [n_nodes,3]
            dipoles.append(node_dipoles)

        # Compute the dipoles
        contributions_dipoles = torch.stack(
            dipoles, dim=-1
        )  # [n_nodes,3,n_contributions]
        atomic_dipoles = torch.sum(contributions_dipoles, dim=-1)  # [n_nodes,3]
        total_dipole = scatter_sum(
            src=atomic_dipoles,
            index=data["batch"],
            dim=0,
            dim_size=num_graphs,
        )  # [n_graphs,3]
        baseline = compute_fixed_charge_dipole(
            charges=data["charges"],
            positions=data["positions"],
            batch=data["batch"],
            num_graphs=num_graphs,
        )  # [n_graphs,3]
        total_dipole = total_dipole + baseline

        output = {
            "dipole": total_dipole,
            "atomic_dipoles": atomic_dipoles,
        }
        return output


@compile_mode("script")
class AtomicDielectricMACE(torch.nn.Module):
    def __init__(
        self,
        r_max: float,
        num_bessel: int,
        num_polynomial_cutoff: int,
        max_ell: int,
        interaction_cls: Type[InteractionBlock],
        interaction_cls_first: Type[InteractionBlock],
        num_interactions: int,
        num_elements: int,
        hidden_irreps: o3.Irreps,
        MLP_irreps: o3.Irreps,
        avg_num_neighbors: float,
        atomic_numbers: List[int],
        correlation: int,
        gate: Optional[Callable],
        atomic_energies: Optional[
            None
        ],  # Just here to make it compatible with energy models, MUST be None
        apply_cutoff: bool = True,  # pylint: disable=unused-argument
        use_reduced_cg: bool = True,  # pylint: disable=unused-argument
        use_so3: bool = False,  # pylint: disable=unused-argument
        distance_transform: str = "None",  # pylint: disable=unused-argument
        radial_type: Optional[str] = "bessel",
        radial_MLP: Optional[List[int]] = None,
        cueq_config: Optional[Dict[str, Any]] = None,  # pylint: disable=unused-argument
        oeq_config: Optional[Dict[str, Any]] = None,  # pylint: disable=unused-argument
        edge_irreps: Optional[o3.Irreps] = None,  # pylint: disable=unused-argument
        dipole_only: Optional[bool] = True,  # pylint: disable=unused-argument
        use_polarizability: Optional[bool] = True,  # pylint: disable=unused-argument
        means_stds: Optional[Dict[str, torch.Tensor]] = None,  # pylint: disable=W0613
    ):
        super().__init__()
        self.register_buffer(
            "atomic_numbers", torch.tensor(atomic_numbers, dtype=torch.int64)
        )
        self.register_buffer("r_max", torch.tensor(r_max, dtype=torch.float64))
        self.register_buffer(
            "num_interactions", torch.tensor(num_interactions, dtype=torch.int64)
        )

        # Predefine buffers to be TorchScript-safe
        self.register_buffer("dipole_mean", torch.zeros(3))
        self.register_buffer("dipole_std", torch.ones(3))
        self.register_buffer(
            "polarizability_mean", torch.zeros(3, 3)
        )  # 3x3 matrix flattened
        self.use_polarizability = use_polarizability
        self.register_buffer("polarizability_std", torch.ones(3, 3))
        self.register_buffer("change_of_basis", get_change_of_basis())
        # self.register_buffer("mean_polarizability_sh", torch.zeros(6))
        # self.register_buffer("std_polarizability_sh", torch.ones(6))
        if means_stds is not None:
            if "dipole_mean" in means_stds:
                self.dipole_mean.data.copy_(means_stds["dipole_mean"])
            if "dipole_std" in means_stds:
                self.dipole_std.data.copy_(means_stds["dipole_std"])
            if "polarizability_mean" in means_stds:
                self.polarizability_mean.data.copy_(means_stds["polarizability_mean"])
            if "polarizability_std" in means_stds:
                self.polarizability_std.data.copy_(means_stds["polarizability_std"])
            # if "mean_polarizability_sh" in means_stds:
            #    self.mean_polarizability_sh.data.copy_(means_stds["mean_polarizability_sh"])
            # if "std_polarizability_sh" in means_stds:
            #    self.std_polarizability_sh.data.copy_(means_stds["std_polarizability_sh"])'''
        assert atomic_energies is None
        # self.use_polarizability = use_polarizability
        # self.use_dipole = use_dipole

        # Embedding
        node_attr_irreps = o3.Irreps([(num_elements, (0, 1))])
        node_feats_irreps = o3.Irreps([(hidden_irreps.count(o3.Irrep(0, 1)), (0, 1))])
        self.node_embedding = LinearNodeEmbeddingBlock(
            irreps_in=node_attr_irreps, irreps_out=node_feats_irreps
        )
        self.radial_embedding = RadialEmbeddingBlock(
            r_max=r_max,
            num_bessel=num_bessel,
            num_polynomial_cutoff=num_polynomial_cutoff,
            radial_type=radial_type,
        )
        edge_feats_irreps = o3.Irreps(f"{self.radial_embedding.out_dim}x0e")

        sh_irreps = o3.Irreps.spherical_harmonics(max_ell)
        num_features = hidden_irreps.count(o3.Irrep(0, 1))
        interaction_irreps = (sh_irreps * num_features).sort()[0].simplify()
        self.spherical_harmonics = o3.SphericalHarmonics(
            sh_irreps, normalize=True, normalization="component"
        )
        if radial_MLP is None:
            radial_MLP = [64, 64, 64]

        # Interactions and readouts
        inter = interaction_cls_first(
            node_attrs_irreps=node_attr_irreps,
            node_feats_irreps=node_feats_irreps,
            edge_attrs_irreps=sh_irreps,
            edge_feats_irreps=edge_feats_irreps,
            target_irreps=interaction_irreps,
            hidden_irreps=hidden_irreps,
            avg_num_neighbors=avg_num_neighbors,
            radial_MLP=radial_MLP,
        )
        self.interactions = torch.nn.ModuleList([inter])

        # Use the appropriate self connection at the first layer
        use_sc_first = False
        if "Residual" in str(interaction_cls_first):
            use_sc_first = True

        node_feats_irreps_out = inter.target_irreps
        prod = EquivariantProductBasisBlock(
            node_feats_irreps=node_feats_irreps_out,
            target_irreps=hidden_irreps,
            correlation=correlation,
            num_elements=num_elements,
            use_sc=use_sc_first,
        )
        self.products = torch.nn.ModuleList([prod])

        self.readouts = torch.nn.ModuleList()
        self.readouts.append(
            LinearDipolePolarReadoutBlock(hidden_irreps, use_polarizability=True)
        )

        for i in range(num_interactions - 1):
            if i == num_interactions - 2:
                # does it always do polar and dipole together?
                assert (
                    len(hidden_irreps) > 1
                ), "To predict dipoles use at least l=1 hidden_irreps"
                # hidden_irreps_out = str(
                #     hidden_irreps[1]
                # )  # Select only l=1 vectors for last layer
                hidden_irreps_out = (
                    hidden_irreps  # this is different in the AtomicDipoleMACE
                )
            else:
                hidden_irreps_out = hidden_irreps
            inter = interaction_cls(
                node_attrs_irreps=node_attr_irreps,
                node_feats_irreps=hidden_irreps,
                edge_attrs_irreps=sh_irreps,
                edge_feats_irreps=edge_feats_irreps,
                target_irreps=interaction_irreps,
                hidden_irreps=hidden_irreps_out,
                avg_num_neighbors=avg_num_neighbors,
                radial_MLP=radial_MLP,
            )
            self.interactions.append(inter)
            prod = EquivariantProductBasisBlock(
                node_feats_irreps=interaction_irreps,
                target_irreps=hidden_irreps_out,
                correlation=correlation,
                num_elements=num_elements,
                use_sc=True,
            )
            self.products.append(prod)
            if i == num_interactions - 2:
                self.readouts.append(
                    NonLinearDipolePolarReadoutBlock(
                        hidden_irreps_out,
                        MLP_irreps,
                        gate,
                        use_polarizability=True,
                    )
                )
                # print("Nonlinear irrpes: ", hidden_irreps_out, MLP_irreps)
                # exit()
            else:
                self.readouts.append(
                    LinearDipolePolarReadoutBlock(
                        hidden_irreps,
                        # use_charge=True,
                        use_polarizability=True,
                    )
                )

    def forward(
        self,
        data: Dict[str, torch.Tensor],
        training: bool = False,  # pylint: disable=W0613
        compute_force: bool = False,
        compute_virials: bool = False,
        compute_stress: bool = False,
        compute_displacement: bool = False,
        compute_dielectric_derivatives: bool = False,  # no training on derivatives
        compute_edge_forces: bool = False,  # pylint: disable=W0613
        compute_atomic_stresses: bool = False,  # pylint: disable=W0613
    ) -> Dict[str, Optional[torch.Tensor]]:
        assert compute_force is False
        assert compute_virials is False
        assert compute_stress is False
        assert compute_displacement is False
        # Setup
        data["node_attrs"].requires_grad_(True)
        data["positions"].requires_grad_(True)
        num_graphs = data["ptr"].numel() - 1
        num_atoms = data["ptr"][1:] - data["ptr"][:-1]

        # Embeddings
        node_feats = self.node_embedding(data["node_attrs"])
        vectors, lengths = get_edge_vectors_and_lengths(
            positions=data["positions"],
            edge_index=data["edge_index"],
            shifts=data["shifts"],
        )
        edge_attrs = self.spherical_harmonics(vectors)
        edge_feats, cutoff = self.radial_embedding(
            lengths, data["node_attrs"], data["edge_index"], self.atomic_numbers
        )

        # Interactions
        charges = []
        dipoles = []
        polarizabilities = []
        for interaction, product, readout in zip(
            self.interactions, self.products, self.readouts
        ):
            node_feats, sc = interaction(
                node_attrs=data["node_attrs"],
                node_feats=node_feats,
                edge_attrs=edge_attrs,
                edge_feats=edge_feats,
                edge_index=data["edge_index"],
                cutoff=cutoff,
            )

            node_feats = product(
                node_feats=node_feats,
                sc=sc,
                node_attrs=data["node_attrs"],
            )

            node_out = readout(node_feats).squeeze(-1)  # [n_nodes,3]
            charges.append(node_out[:, 0])

            if self.use_polarizability:
                node_dipoles = node_out[:, 2:5]
                node_polarizability = torch.cat(
                    (node_out[:, 1].unsqueeze(-1), node_out[:, 5:]), dim=-1
                )
                polarizabilities.append(node_polarizability)
                dipoles.append(node_dipoles)
            else:
                raise ValueError(
                    "Polarizability is not used in this model, but it is required for the AtomicDielectricMACE."
                )
        contributions_dipoles = torch.stack(
            dipoles, dim=-1
        )  # [n_nodes,3,n_contributions]
        atomic_dipoles = torch.sum(contributions_dipoles, dim=-1)  # [n_nodes,3]
        atomic_charges = torch.stack(charges, dim=-1).sum(-1)  # [n_nodes,]
        # The idea is to normalize the charges so that they sum to the net charge in the system before predicting the dipole.
        total_charge_excess = scatter_mean(
            src=atomic_charges, index=data["batch"], dim_size=num_graphs
        ) - (data["total_charge"] / num_atoms)
        atomic_charges = atomic_charges - total_charge_excess[data["batch"]]
        total_dipole = scatter_sum(
            src=atomic_dipoles,
            index=data["batch"],
            dim=0,
            dim_size=num_graphs,
        )  # [n_graphs,3]
        baseline = compute_fixed_charge_dipole_polar(
            charges=atomic_charges,  # or data["charges"], ?????
            positions=data["positions"],
            batch=data["batch"],
            num_graphs=num_graphs,
        )  # [n_graphs,3]
        total_dipole = total_dipole + baseline

        if self.use_polarizability:
            # Compute the polarizabilities
            contributions_polarizabilities = torch.stack(
                polarizabilities, dim=-1
            )  # [n_nodes,6,n_contributions]
            atomic_polarizabilities = torch.sum(
                contributions_polarizabilities, dim=-1
            )  # [n_nodes,6]
            total_polarizability_spherical = scatter_sum(
                src=atomic_polarizabilities,
                index=data["batch"],
                dim=0,
                dim_size=num_graphs,
            )  # [n_graphs,6]
            total_polarizability = spherical_to_cartesian(
                total_polarizability_spherical, self.change_of_basis
            )

            if compute_dielectric_derivatives:
                dmu_dr = compute_dielectric_gradients(
                    dielectric=total_dipole,
                    positions=data["positions"],
                )
                dalpha_dr = compute_dielectric_gradients(
                    dielectric=total_polarizability.flatten(-2),
                    positions=data["positions"],
                )
            else:
                dmu_dr = None
                dalpha_dr = None
        else:
            if compute_dielectric_derivatives:
                dmu_dr = compute_dielectric_gradients(
                    dielectric=total_dipole,
                    positions=data["positions"],
                )
            else:
                dmu_dr = None
            total_polarizability = None
            total_polarizability_spherical = None
            dalpha_dr = None

        output = {
            "charges": atomic_charges,
            "dipole": total_dipole,
            "atomic_dipoles": atomic_dipoles,
            "polarizability": total_polarizability,
            "polarizability_sh": total_polarizability_spherical,
            "dmu_dr": dmu_dr,
            "dalpha_dr": dalpha_dr,
        }
        return output


@compile_mode("script")
class EnergyDipolesMACE(torch.nn.Module):
    def __init__(
        self,
        r_max: float,
        num_bessel: int,
        num_polynomial_cutoff: int,
        max_ell: int,
        interaction_cls: Type[InteractionBlock],
        interaction_cls_first: Type[InteractionBlock],
        num_interactions: int,
        num_elements: int,
        hidden_irreps: o3.Irreps,
        MLP_irreps: o3.Irreps,
        avg_num_neighbors: float,
        atomic_numbers: List[int],
        correlation: int,
        gate: Optional[Callable],
        atomic_energies: Optional[np.ndarray],
        apply_cutoff: bool = True,  # pylint: disable=unused-argument
        use_reduced_cg: bool = True,  # pylint: disable=unused-argument
        use_so3: bool = False,  # pylint: disable=unused-argument
        distance_transform: str = "None",  # pylint: disable=unused-argument
        radial_MLP: Optional[List[int]] = None,
        cueq_config: Optional[Dict[str, Any]] = None,  # pylint: disable=unused-argument
        oeq_config: Optional[Dict[str, Any]] = None,  # pylint: disable=unused-argument
        edge_irreps: Optional[o3.Irreps] = None,  # pylint: disable=unused-argument
    ):
        super().__init__()
        self.register_buffer(
            "atomic_numbers", torch.tensor(atomic_numbers, dtype=torch.int64)
        )
        self.register_buffer("r_max", torch.tensor(r_max, dtype=torch.float64))
        self.register_buffer(
            "num_interactions", torch.tensor(num_interactions, dtype=torch.int64)
        )
        # Embedding
        node_attr_irreps = o3.Irreps([(num_elements, (0, 1))])
        node_feats_irreps = o3.Irreps([(hidden_irreps.count(o3.Irrep(0, 1)), (0, 1))])
        self.node_embedding = LinearNodeEmbeddingBlock(
            irreps_in=node_attr_irreps, irreps_out=node_feats_irreps
        )
        self.radial_embedding = RadialEmbeddingBlock(
            r_max=r_max,
            num_bessel=num_bessel,
            num_polynomial_cutoff=num_polynomial_cutoff,
        )
        edge_feats_irreps = o3.Irreps(f"{self.radial_embedding.out_dim}x0e")

        sh_irreps = o3.Irreps.spherical_harmonics(max_ell)
        num_features = hidden_irreps.count(o3.Irrep(0, 1))
        interaction_irreps = (sh_irreps * num_features).sort()[0].simplify()
        self.spherical_harmonics = o3.SphericalHarmonics(
            sh_irreps, normalize=True, normalization="component"
        )
        if radial_MLP is None:
            radial_MLP = [64, 64, 64]
        # Interactions and readouts
        self.atomic_energies_fn = AtomicEnergiesBlock(atomic_energies)

        inter = interaction_cls_first(
            node_attrs_irreps=node_attr_irreps,
            node_feats_irreps=node_feats_irreps,
            edge_attrs_irreps=sh_irreps,
            edge_feats_irreps=edge_feats_irreps,
            target_irreps=interaction_irreps,
            hidden_irreps=hidden_irreps,
            avg_num_neighbors=avg_num_neighbors,
            radial_MLP=radial_MLP,
        )
        self.interactions = torch.nn.ModuleList([inter])

        # Use the appropriate self connection at the first layer
        use_sc_first = False
        if "Residual" in str(interaction_cls_first):
            use_sc_first = True

        node_feats_irreps_out = inter.target_irreps
        prod = EquivariantProductBasisBlock(
            node_feats_irreps=node_feats_irreps_out,
            target_irreps=hidden_irreps,
            correlation=correlation,
            num_elements=num_elements,
            use_sc=use_sc_first,
        )
        self.products = torch.nn.ModuleList([prod])

        self.readouts = torch.nn.ModuleList()
        self.readouts.append(LinearDipoleReadoutBlock(hidden_irreps, dipole_only=False))

        for i in range(num_interactions - 1):
            if i == num_interactions - 2:
                assert (
                    len(hidden_irreps) > 1
                ), "To predict dipoles use at least l=1 hidden_irreps"
                hidden_irreps_out = str(
                    hidden_irreps[:2]
                )  # Select scalars and l=1 vectors for last layer
            else:
                hidden_irreps_out = hidden_irreps
            inter = interaction_cls(
                node_attrs_irreps=node_attr_irreps,
                node_feats_irreps=hidden_irreps,
                edge_attrs_irreps=sh_irreps,
                edge_feats_irreps=edge_feats_irreps,
                target_irreps=interaction_irreps,
                hidden_irreps=hidden_irreps_out,
                avg_num_neighbors=avg_num_neighbors,
                radial_MLP=radial_MLP,
            )
            self.interactions.append(inter)
            prod = EquivariantProductBasisBlock(
                node_feats_irreps=interaction_irreps,
                target_irreps=hidden_irreps_out,
                correlation=correlation,
                num_elements=num_elements,
                use_sc=True,
            )
            self.products.append(prod)
            if i == num_interactions - 2:
                self.readouts.append(
                    NonLinearDipoleReadoutBlock(
                        hidden_irreps_out, MLP_irreps, gate, dipole_only=False
                    )
                )
            else:
                self.readouts.append(
                    LinearDipoleReadoutBlock(hidden_irreps, dipole_only=False)
                )

    def forward(
        self,
        data: Dict[str, torch.Tensor],
        training: bool = False,
        compute_force: bool = True,
        compute_virials: bool = False,
        compute_stress: bool = False,
        compute_displacement: bool = False,
        compute_edge_forces: bool = False,  # pylint: disable=W0613
        compute_atomic_stresses: bool = False,  # pylint: disable=W0613
    ) -> Dict[str, Optional[torch.Tensor]]:
        # Setup
        data["node_attrs"].requires_grad_(True)
        data["positions"].requires_grad_(True)
        num_graphs = data["ptr"].numel() - 1
        num_atoms_arange = torch.arange(data["positions"].shape[0])
        displacement = torch.zeros(
            (num_graphs, 3, 3),
            dtype=data["positions"].dtype,
            device=data["positions"].device,
        )
        if compute_virials or compute_stress or compute_displacement:
            (
                data["positions"],
                data["shifts"],
                displacement,
            ) = get_symmetric_displacement(
                positions=data["positions"],
                unit_shifts=data["unit_shifts"],
                cell=data["cell"],
                edge_index=data["edge_index"],
                num_graphs=num_graphs,
                batch=data["batch"],
            )

        # Atomic energies
        node_e0 = self.atomic_energies_fn(data["node_attrs"])[
            num_atoms_arange, data["head"][data["batch"]]
        ]
        e0 = scatter_sum(
            src=node_e0, index=data["batch"], dim=-1, dim_size=num_graphs
        )  # [n_graphs,]

        # Embeddings
        node_feats = self.node_embedding(data["node_attrs"])
        vectors, lengths = get_edge_vectors_and_lengths(
            positions=data["positions"],
            edge_index=data["edge_index"],
            shifts=data["shifts"],
        )
        edge_attrs = self.spherical_harmonics(vectors)
        edge_feats, cutoff = self.radial_embedding(
            lengths, data["node_attrs"], data["edge_index"], self.atomic_numbers
        )

        # Interactions
        energies = [e0]
        node_energies_list = [node_e0]
        dipoles = []
        for interaction, product, readout in zip(
            self.interactions, self.products, self.readouts
        ):
            node_feats, sc = interaction(
                node_attrs=data["node_attrs"],
                node_feats=node_feats,
                edge_attrs=edge_attrs,
                edge_feats=edge_feats,
                edge_index=data["edge_index"],
                cutoff=cutoff,
            )
            node_feats = product(
                node_feats=node_feats,
                sc=sc,
                node_attrs=data["node_attrs"],
            )
            node_out = readout(node_feats).squeeze(-1)  # [n_nodes, ]
            # node_energies = readout(node_feats).squeeze(-1)  # [n_nodes, ]
            node_energies = node_out[:, 0]
            energy = scatter_sum(
                src=node_energies, index=data["batch"], dim=-1, dim_size=num_graphs
            )  # [n_graphs,]
            energies.append(energy)
            node_dipoles = node_out[:, 1:]
            dipoles.append(node_dipoles)

        # Compute the energies and dipoles
        contributions = torch.stack(energies, dim=-1)
        total_energy = torch.sum(contributions, dim=-1)  # [n_graphs, ]
        node_energy_contributions = torch.stack(node_energies_list, dim=-1)
        node_energy = torch.sum(node_energy_contributions, dim=-1)  # [n_nodes, ]
        contributions_dipoles = torch.stack(
            dipoles, dim=-1
        )  # [n_nodes,3,n_contributions]
        atomic_dipoles = torch.sum(contributions_dipoles, dim=-1)  # [n_nodes,3]
        total_dipole = scatter_sum(
            src=atomic_dipoles,
            index=data["batch"].unsqueeze(-1),
            dim=0,
            dim_size=num_graphs,
        )  # [n_graphs,3]
        baseline = compute_fixed_charge_dipole(
            charges=data["charges"],
            positions=data["positions"],
            batch=data["batch"],
            num_graphs=num_graphs,
        )  # [n_graphs,3]
        total_dipole = total_dipole + baseline

        forces, virials, stress, _, _ = get_outputs(
            energy=total_energy,
            positions=data["positions"],
            displacement=displacement,
            cell=data["cell"],
            training=training,
            compute_force=compute_force,
            compute_virials=compute_virials,
            compute_stress=compute_stress,
        )

        output = {
            "energy": total_energy,
            "node_energy": node_energy,
            "contributions": contributions,
            "forces": forces,
            "virials": virials,
            "stress": stress,
            "displacement": displacement,
            "dipole": total_dipole,
            "atomic_dipoles": atomic_dipoles,
        }
        return output



"""Input and output functions."""
from __future__ import annotations

from .core import Configuration, MLAB

from collections import Counter, OrderedDict

import itertools
from itertools import chain

import numpy as np

from ase.data import atomic_masses, atomic_numbers
from ase.atoms import Atoms

_LATTICE_FMT_STR = (
    "{:.16f}  {:.16f}  {:.16f}\n{:.16f}  {:.16f}  {:.16f}\n{:.16f}  {:.16f}  {:.16f}"
)
_BASIS_FMT_STR = """**************************************************
     Basis set for {}
--------------------------------------------------
{}"""
_CONFIGURATION_FMT_STR = """**************************************************
     Configuration num.    {}
==================================================
     System name
--------------------------------------------------
     {}
==================================================
     The number of atom types
--------------------------------------------------
       {}
==================================================
     The number of atoms
--------------------------------------------------
         {}
**************************************************
     Atom types and atom numbers
--------------------------------------------------
{}
==================================================
     CTIFOR
--------------------------------------------------
   {:.14f}
==================================================
     Primitive lattice vectors (ang.)
--------------------------------------------------
{}
==================================================
     Atomic positions (ang.)
--------------------------------------------------
{}
==================================================
     Total energy (eV)
--------------------------------------------------
  {:.14f}
==================================================
     Forces (eV ang.^-1)
--------------------------------------------------
{}
==================================================
     Stress (kbar)
--------------------------------------------------
     XX YY ZZ
--------------------------------------------------
 {}
--------------------------------------------------
     XY YZ ZX
--------------------------------------------------
 {}"""
_ML_AB_FMT_STR = """ {}
**************************************************
     The number of configurations
--------------------------------------------------
        {}
**************************************************
     The maximum number of atom type
--------------------------------------------------
       {}
**************************************************
     The atom types in the data file
--------------------------------------------------
     {}
**************************************************
     The maximum number of atoms per system
--------------------------------------------------
            {}
**************************************************
     The maximum number of atoms per atom type
--------------------------------------------------
            {}
**************************************************
     Reference atomic energy (eV)
--------------------------------------------------
   {}
**************************************************
     Atomic mass
--------------------------------------------------
   {}
**************************************************
     The numbers of basis sets per atom type
--------------------------------------------------
       {}
{}
{}
"""
_INIT_STR_EXTYXZ = """energy={} stress="{}" Lattice="{}" pbc="T T T" Properties=species:S:1:pos:R:3:forces:R:3"""
_KBAR_TO_EV_Acub = 6.242e-4  # Convert VASP kbar stress to eV/A^3


def read_ml_ab_file(filename):
    """
    Read an MLAB file.

    Parameters
    ----------
    filename
        Path to an MLAB file.

    Returns
    -------
    MLAB
        An MLAB object.
    """
    from .core import MLAB

    configurations = []
    configuration_data = None
    header_data = []
    header = None
    with open(filename) as file:
        for i, line in enumerate(file):
            if "Configuration num." in line:
                if configuration_data:
                    configurations.append(_parse_config(configuration_data))
                elif configuration_data is None:
                    header = _parse_header(header_data)
                configuration_data = []
            elif configuration_data is None:
                header_data.append(line.rstrip())
            else:
                configuration_data.append(line.rstrip())

        # end of file; parse final config
        configurations.append(_parse_config(configuration_data))

    return MLAB(
        configurations=configurations,
        basis_set=header["basis_set"],
        atomic_mass=header["atomic_mass"],
        reference_energy=header["reference_energy"],
        version=header["version"],
    )


def ml_ab_to_string(ml_ab):
    """
    Convert an MLAB object to a string representation.

    Parameters
    ----------
    ml_ab : MLAB
        An MLAB object.

    Returns
    -------
    str
        The string representation of an MLAB object.
    """
    ref_energy_fmt = _three_fmt(
        map("{:.14f} ".format, ml_ab.reference_energy_per_type), prefix="   "
    )
    atomic_mass_fmt = _three_fmt(
        map("{:.14f} ".format, ml_ab.atomic_mass_per_type), prefix="   "
    )
    basis_number_fmt = _three_fmt(
        map(str, ml_ab.num_basis_set_per_type), prefix="       "
    )
    basis_fmt = "\n".join(
        [
            _BASIS_FMT_STR.format(
                x, "\n".join(["    {}   {}".format(*i) for i in ml_ab.basis_set[x]])
            )
            for x in ml_ab.atom_types
        ]
    )
    configuration_fmt = "\n".join(
        [
            _CONFIGURATION_FMT_STR.format(
                i + 1,
                c.name,
                c.num_atom_types,
                c.num_atoms,
                "\n".join(
                    ["    {}   {}".format(*i) for i in c.atom_types_numbers.items()]
                ),
                c.ctifor,
                _LATTICE_FMT_STR.format(*chain.from_iterable(c.lattice)),
                "\n".join(["{:.16f}   {:.16f}   {:.16f}".format(*i) for i in c.coords]),
                c.total_energy,
                "\n".join(["{:.16f}   {:.16f}   {:.16f}".format(*i) for i in c.forces]),
                "{:.16f}   {:.16f}   {:.16f}".format(*c.stress_diagonal),
                "{:.16f}   {:.16f}   {:.16f}".format(*c.stress_off_diagonal),
            )
            for i, c in enumerate(ml_ab.configurations)
        ]
    )

    return _ML_AB_FMT_STR.format(
        ml_ab.version,
        ml_ab.num_configurations,
        ml_ab.num_atom_types,
        _three_fmt(ml_ab.atom_types, prefix="     "),
        ml_ab.max_num_atoms,
        ml_ab.max_num_atoms_per_type,
        ref_energy_fmt,
        atomic_mass_fmt,
        basis_number_fmt,
        basis_fmt,
        configuration_fmt,
    )


def _parse_config(config):
    from .core import Configuration

    name = config[3]
    natom_types = int(config[7])
    natoms = int(config[11])
    atom_types_numbers = {
        x.split()[0]: int(x.split()[1]) for x in config[15 : 15 + natom_types]
    }
    ctifor = float(config[18 + natom_types])
    lattice = [
        list(map(float, x.split())) for x in config[22 + natom_types : 25 + natom_types]
    ]
    coords = [
        list(map(float, x.split()))
        for x in config[28 + natom_types : 28 + natom_types + natoms]
    ]
    total_energy = float(config[31 + natom_types + natoms])
    forces = [
        list(map(float, x.split()))
        for x in config[35 + natom_types + natoms : 35 + natom_types + natoms * 2]
    ]
    stress_diagonal = list(map(float, config[40 + natom_types + natoms * 2].split()))
    stress_off_diagonal = list(
        map(float, config[44 + natom_types + natoms * 2].split())
    )

    return Configuration(
        name=name,
        atom_types_numbers=atom_types_numbers,
        ctifor=ctifor,
        lattice=np.array(lattice),
        coords=np.array(coords),
        forces=np.array(forces),
        total_energy=total_energy,
        stress_diagonal=np.array(stress_diagonal),
        stress_off_diagonal=np.array(stress_off_diagonal),
    )


def _parse_header(header):
    version = header[0]
    natom_types = int(header[8])
    nlines = int(np.ceil(natom_types / 3))
    atom_types = " ".join(header[12 : 12 + nlines]).split()
    reference_energy = list(
        map(float, " ".join(header[23 + nlines : 23 + nlines * 2]).split())
    )
    atomic_mass = list(
        map(float, " ".join(header[26 + nlines * 2 : 26 + nlines * 3]).split())
    )
    num_basis = list(
        map(int, " ".join(header[29 + nlines * 3 : 29 + nlines * 4]).split())
    )
    basis_set = {}

    for i, nbasis in enumerate(num_basis):
        c = sum(num_basis[:i]) + 3 * i
        basis_set[atom_types[i]] = [
            list(map(int, x.split()))
            for x in header[32 + nlines * 4 + c : 32 + nlines * 4 + nbasis + c]
        ]

    return {
        "version": version,
        "atom_types": atom_types,
        "reference_energy": dict(zip(atom_types, reference_energy)),
        "atomic_mass": dict(zip(atom_types, atomic_mass)),
        "num_basis": num_basis,
        "basis_set": basis_set,
    }


def _grouper(iterable, n):
    """
    Collect data into fixed-length chunks or blocks.

    >>> list(grouper('ABCDEFG', 3))
    [['A', 'B', 'C'], ['D', 'E', 'F'], ['G']]
    """
    iterable = iter(iterable)
    return iter(lambda: list(itertools.islice(iterable, n)), [])


def _three_fmt(obj, prefix=""):
    """Format a list as three items per line."""
    return f"\n{prefix}".join([" ".join(x) for x in _grouper(obj, 3)])


def ml_ab_to_extxyz(ml_ab, filename, stress_unit=None):
    """
    Convert an MLAB object to an extended XYZ string representation.
    Parameters
    ----------
    ml_ab
        An MLAB object.
    filename
        Path to an extended XYZ file.
    stress_unit
        Unit of stress to convert to. VASP units are kbar. Default is None which leaves the units alone.
        If 'ev/A^3', the stress is converted from kbar to eV/A^3 (with a negative sign convention assumed).
    """
    if stress_unit == "kbar" or not stress_unit:
        stress_unit = 1
    elif stress_unit == "eV/A^3":
        stress_unit = -1 * _KBAR_TO_EV_Acub
    else:
        raise ValueError(f"Unknown stress unit: {stress_unit}")
    with open(filename, "w") as f:
        for i, conf in enumerate(ml_ab.configurations):
            stress = stress_unit * np.array(
                [
                    conf.stress_diagonal[0],
                    conf.stress_off_diagonal[0],
                    conf.stress_off_diagonal[2],
                    conf.stress_off_diagonal[0],
                    conf.stress_diagonal[1],
                    conf.stress_off_diagonal[1],
                    conf.stress_off_diagonal[2],
                    conf.stress_off_diagonal[1],
                    conf.stress_diagonal[2],
                ]
            )
            f.write(str(conf.num_atoms) + "\n")

            f.write(
                _INIT_STR_EXTYXZ.format(
                    conf.total_energy,
                    " ".join(map("{:.16f}".format, stress)),
                    " ".join(map("{:.16f}".format, conf.lattice.flatten())),
                )
                + "\n"
            )
            c = 0
            for el, num in conf.atom_types_numbers.items():
                for j in range(num):
                    f.write(el + "         ")
                    f.write(" ".join(map("{:.16f}".format, conf.coords[c, :])) + " ")
                    f.write(" ".join(map("{:.16f}".format, conf.forces[c, :])) + "\n")
                    c += 1
        f.close()


def config_from_atoms(a) -> Configuration:
    """
    Transforms an ASE Atoms object into a Configuration object.

    Parameters
    ----------
    a (ase.atoms.Atoms)
        ASE Atoms object

    Returns
    -------
    Configuration
        Configuration object
    """
    _EV_Acub_TO_KBAR = -1 / _KBAR_TO_EV_Acub # From eV/A^3 to kbar. Assumes negative sign convention in VASP
    xx, yy, zz, yz, xz, xy = a.get_stress(voigt=True)
    config = Configuration(
        name=a.get_chemical_formula(),
        atom_types_numbers=dict(Counter(a.get_chemical_symbols())), # eg {'Cd': 32, 'Te': 32}
        ctifor=0.001, # Needs to be set for writting MLAB with pymlff, so spurious value.
        # If desired to overwrite, can set ISTART=3 and CTIFOR=0.01 in INCAR.
        # (see https://www.vasp.at/forum/viewtopic.php?t=18400.
        lattice=a.get_cell().array,
        coords=a.get_positions(),
        forces=a.get_forces(),
        total_energy=a.calc.results["free_energy"], # VASP MLFF reads the potential E + electronic entropy.
        # See https://w.vasp.at/forum/viewtopic.php?t=13705
        stress_diagonal=_EV_Acub_TO_KBAR * np.array([xx, yy, zz]),
        stress_off_diagonal=_EV_Acub_TO_KBAR * np.array([xy, yz, xz]),
    )
    return config


def ml_ab_from_trajectory(
    trajectory: list[Atoms],
    basis_set: dict[str: [tuple[int, int],]]=None,
    version: str=' 1.0 Version'
) -> MLAB:
    """
    Transforms an ASE trajectory into a MLAB object.

    Parameters
    ----------
    trajectory: list[Atoms]
        List of ASE Atoms objects.
    basis_set: dict
        The basis set for the trained force field. It is formatted as a dictionary mapping
        the elemental species to the basis set. The basis set is specified as a list of
        [configuration_index, atom_index], that specifies which atomic environments are
        included in the training set for each species. Note that the configuration and
        atom index are 1-indexed as required by VASP. If not set, will default to [(1,1),]
        for each element, and should be combined with `ML_ISTART=3`, as explained here:
        https://www.vasp.at/forum/viewtopic.php?t=18400
    version: str
        The version of the ML_AB file format. Defaults to ' 1.0 Version'.

    Returns
    -------
    MLAB
        MLAB object
    """
    symbols = [list(OrderedDict.fromkeys(a.get_chemical_symbols())) for a in trajectory]
    symbols_unique = list(OrderedDict.fromkeys(itertools.chain.from_iterable(symbols)))
    atomic_mass = {
        species: atomic_masses[atomic_numbers[species]] for species in symbols_unique
    }
    configs = [config_from_atoms(a) for a in trajectory]
    reference_energy = {s: 0.0 for s in symbols_unique}
    if not basis_set:
        basis_set = {s: [[1, 1], ] for s in symbols_unique}
    return  MLAB(
        configurations=configs,
        basis_set=basis_set,
        atomic_mass=atomic_mass,
        reference_energy=reference_energy,
        version=version,
    )



"""Utility functions for introducing zero force constants."""

import itertools
from typing import Union

import numpy as np
import macer.externals.spglib_bundled as spglib

try:
    spglib.error.OLD_ERROR_HANDLING = False
except AttributeError:
    pass

from macer.externals.symfc_bundled.utils.utils import SymfcAtoms


def apply_zeros(C, zero_ids):
    """Assign zero to matrix elements.

    Use this function, sparse C can become larger by assigning zeros. Zero
    elements should be applied to c_trans and c_perm in constructing them.
    Warning: Slow when zero_ids is large.

    Method 1
    --------
    C[zero_ids,:] = 0

    Method 2
    --------
    C = C.tolil() C[zero_ids, :] = 0 C = C.tocsr()

    """
    for i in zero_ids:
        nonzero_cols = C.getrow(i).nonzero()[1]
        for j in nonzero_cols:
            C[i, j] = 0
    return C


class FCCutoff:
    """Class for introducing cutoff radius."""

    def __init__(self, supercell: SymfcAtoms, cutoff: float = 7.0):
        """Find nonzero FC elements given by cutoff distance.

        For FC3, zero elements are determined by the distance between three
        pairs of atoms.

        Parameters
        ----------
        supercell: SymfcAtoms
            Supercell structure.
        cutoff: float
            Cutoff distance between two atoms.

        """
        self._supercell = supercell
        self._cutoff = cutoff
        self._n_atom = supercell.scaled_positions.shape[0]

        self._distances = self._calc_distances()
        self._neighbors = None
        self._nonzero_fc2 = None
        self._nonzero_fc3 = None
        self._nonzero_fc4 = None

    @property
    def neighbors(self) -> list[np.ndarray]:
        """Neighbor atoms: shape=(n_atom, n_neighbors)."""
        if self._neighbors is None:
            self._neighbors = [
                np.where(self._distances[i] < self._cutoff)[0]
                for i in range(self._n_atom)
            ]
        return self._neighbors

    @property
    def outsides(self) -> list[np.ndarray]:
        """Atoms outside cutoff radius: shape=(n_atom, n_neighbors)."""
        return [
            np.where(self._distances[i] >= self._cutoff)[0] for i in range(self._n_atom)
        ]

    @property
    def distances(self) -> np.ndarray:
        """Minimum distances between atoms: shape=(n_atom, n_atom)."""
        return self._distances

    def combinations1(self) -> np.ndarray:
        """Return combinations with single index ia."""
        return np.array([[i] for i in range(3 * self._n_atom)], dtype=int)

    def combinations2(self) -> np.ndarray:
        """Return combinations with two distinguished indices (ia,jb)."""
        combinations = []
        for jb in range(3 * self._n_atom):
            j = jb // 3
            combs = [
                [3 * i + a, jb]
                for i in self.neighbors[j]
                for a in range(3)
                if 3 * i + a < jb
            ]
            combinations.extend(combs)
        combinations_fc2 = np.array(combinations)
        return combinations_fc2

    def combinations3_all(self) -> np.ndarray:
        """Return combinations with three distinguished indices (ia,jb,kc)."""
        combinations = []
        for kc in range(3 * self._n_atom):
            combs = self.combinations3(kc)
            combinations.extend(combs)
        combinations_fc3 = np.array(combinations)
        return combinations_fc3

    def combinations3(self, kc) -> Union[list, np.ndarray]:
        """Return combinations with three distinguished indices (ia,jb,kc).

        Return only combinations with kc.
        """
        k = kc // 3
        neighbors_N3 = [
            3 * j + b for j in self.neighbors[k] for b in range(3) if 3 * j + b < kc
        ]
        combs = np.array(list(itertools.combinations(neighbors_N3, 2)))
        if len(combs) > 0:
            indices = np.where(
                self.distances[(combs[:, 0] // 3, combs[:, 1] // 3)] < self._cutoff
            )[0]
            combs = combs[indices]
            return np.hstack([combs, np.full((combs.shape[0], 1), kc)])
        return []

    def combinations4_all(self) -> np.ndarray:
        """Return combinations with three distinguished indices (ia,jb,kc,ld)."""
        combinations = []
        for ld in range(3 * self._n_atom):
            combs = self.combinations4(ld)
            combinations.extend(combs)
        combinations_fc4 = np.array(combinations)
        return combinations_fc4

    def combinations4(self, ld) -> Union[list, np.ndarray]:
        """Return combinations with three distinguished indices (ia,jb,kc,ld).

        Return only combinations with kc.

        """
        ll = ld // 3
        neighbors_N3 = [
            3 * j + b for j in self.neighbors[ll] for b in range(3) if 3 * j + b < ld
        ]
        combs = np.array(list(itertools.combinations(neighbors_N3, 3)))
        if len(combs) > 0:
            indices = np.where(
                (self.distances[(combs[:, 0] // 3, combs[:, 1] // 3)] < self._cutoff)
                & (self.distances[(combs[:, 0] // 3, combs[:, 2] // 3)] < self._cutoff)
                & (self.distances[(combs[:, 1] // 3, combs[:, 2] // 3)] < self._cutoff)
            )[0]
            combs = combs[indices]
            return np.hstack([combs, np.full((combs.shape[0], 1), ld)])
        return []

    def nonzero_atomic_indices_fc2(self) -> np.ndarray:
        """Return atomic indices of nonzero FC2.

        Returns
        -------
        nonzero : np.ndarray
            FC2 element is nonzero (True) or zero (False), shape=(NN).

        """
        if self._nonzero_fc2 is not None:
            return self._nonzero_fc2

        self._nonzero_fc2 = nonzero = np.zeros(self._n_atom**2, dtype=bool)
        for i in range(self._n_atom):
            ids = np.array(self.neighbors[i]) + i * self._n_atom
            nonzero[ids] = True
        return self._nonzero_fc2

    def nonzero_atomic_indices_fc3(self) -> np.ndarray:
        """Return atomic indices of nonzero FC3.

        Returns
        -------
        nonzero : np.ndarray
            FC3 element is nonzero (True) or zero (False), shape=(NNN).
        """
        if self._nonzero_fc3 is not None:
            return self._nonzero_fc3

        self._nonzero_fc3 = nonzero = np.zeros(self._n_atom**3, dtype=bool)
        for i in range(self._n_atom):
            jlist = self.neighbors[i]
            combs = np.array(list(itertools.product(jlist, jlist)))
            if len(combs) > 0:
                combs = combs[self.distances[(combs[:, 0], combs[:, 1])] < self._cutoff]
                ids = combs @ np.array([self._n_atom, 1]) + i * self._n_atom**2
                nonzero[ids] = True
        return self._nonzero_fc3

    def nonzero_atomic_indices_fc4(self) -> np.ndarray:
        """Return atomic indices of nonzero FC4.

        Returns
        -------
        nonzero : np.ndarray
            FC4 element is nonzero (True) or zero (False), shape=(NNNN).

        """
        if self._nonzero_fc4 is not None:
            return self._nonzero_fc4

        self._nonzero_fc4 = nonzero = np.zeros(self._n_atom**4, dtype=bool)
        for i in range(self._n_atom):
            jlist = self.neighbors[i]
            combs = np.array(list(itertools.product(*[jlist, jlist, jlist])))
            if len(combs) > 0:
                combs = combs[
                    (self.distances[(combs[:, 0], combs[:, 1])] < self._cutoff)
                    & (self.distances[(combs[:, 0], combs[:, 2])] < self._cutoff)
                    & (self.distances[(combs[:, 1], combs[:, 2])] < self._cutoff)
                ]
                ids = combs @ np.array([self._n_atom**2, self._n_atom, 1])
                ids += i * self._n_atom**3
                nonzero[ids] = True
        return self._nonzero_fc4

    def _calc_distances(self) -> np.ndarray:
        """Calculate minimum distances between atoms.

        This algorithm must be reconsidered.
        (Not applicable to structures with strange lattice shape)

        """
        reduced_bases = spglib.niggli_reduce(self._supercell.cell)  # type: ignore
        trans_mat_float = self._supercell.cell @ np.linalg.inv(reduced_bases)  # type: ignore
        trans_mat = np.rint(trans_mat_float).astype(int)
        assert (np.abs(trans_mat_float - trans_mat) < 1e-8).all()

        scaled_positions = self._supercell.scaled_positions @ trans_mat
        scaled_positions -= np.rint(scaled_positions)
        diff = scaled_positions[:, None, :] - scaled_positions[None, :, :]

        trans = np.array(list(itertools.product(*[[-1, 0, 1], [-1, 0, 1], [-1, 0, 1]])))
        norms = np.ones((self._n_atom, self._n_atom)) * 1e10
        for t1 in trans:
            t1_tile = np.tile(t1, (self._n_atom, self._n_atom, 1))
            norms_trial = np.linalg.norm((diff - t1_tile) @ reduced_bases, axis=2)
            match = norms_trial < norms
            norms[match] = norms_trial[match]
        self._distances = norms
        return self._distances



"""
Macer: Machine-learning accelerated Atomic Computational Environment for automated Research workflows
Copyright (c) 2025 The Macer Package Authors
Author: Soungmin Bae <soungminbae@gmail.com>
License: MIT
"""

import sys
from ase.calculators.emt import EMT

_EMT_AVAILABLE = True
EMT_SUPPORTED_ELEMENTS = ["Al", "Cu", "Ag", "Au", "Ni", "Pd", "Pt"]

def get_emt_calculator(**kwargs):
    """
    Returns an ASE EMT calculator instance.
    Includes validation for supported elements.
    """
    print(f"EMT calculator selected. Supported elements: {', '.join(EMT_SUPPORTED_ELEMENTS)}")
    return EMT()

def validate_emt_elements(atoms):
    """
    Checks if all atoms in the object are supported by EMT.
    Exits if unsupported elements are found.
    """
    symbols = set(atoms.get_chemical_symbols())
    unsupported = [s for s in symbols if s not in EMT_SUPPORTED_ELEMENTS]
    
    if unsupported:
        print(f"\nError: EMT does not support the following elements: {', '.join(unsupported)}")
        print(f"Supported elements are: {', '.join(EMT_SUPPORTED_ELEMENTS)}")
        sys.exit(1)



#!/usr/bin/env python3
from __future__ import annotations

import argparse
from pathlib import Path


def parse_parameters(path: Path) -> dict[str, str]:
    data: dict[str, str] = {}
    for line in path.read_text().splitlines():
        line = line.strip()
        if not line or line.startswith("#") or "=" not in line:
            continue
        k, v = line.split("=", 1)
        data[k.strip()] = v.strip()
    return data


def map_to_macer_cli(params: dict[str, str]) -> list[str]:
    mapped = []
    if "T_K" in params:
        mapped += ["-T", params["T_K"]]
    if "nconf" in params:
        mapped += ["--nconf", params["nconf"]]
    if "nfits" in params:
        mapped += ["--nfits", params["nfits"]]
    if "memory" in params:
        mapped += ["--memory", params["memory"]]
    if "mixing" in params:
        mapped += ["--mixing", params["mixing"]]
    if "tolerance" in params:
        mapped += ["--tolerance", params["tolerance"]]
    if "pdiff" in params:
        mapped += ["--pdiff", params["pdiff"]]
    if "grid" in params:
        mapped += ["--grid", params["grid"]]
    if "imaginary_freq" in params:
        mapped += ["--imaginary-freq", params["imaginary_freq"]]
    if params.get("use_smalldisp", "False").lower() in ("true", "1", "yes"):
        mapped += ["--use-smalldisp"]
    if "use_pressure" in params:
        mapped += ["--use-pressure", params["use_pressure"]]
    if "pressure_diag" in params:
        vals = [x.strip() for x in params["pressure_diag"].split(",")]
        if len(vals) == 3:
            mapped += ["--pressure-diag", vals[0], vals[1], vals[2]]
    return mapped


def main() -> int:
    parser = argparse.ArgumentParser(description="Compare qscaild-master parameter files with macer qscaild option mapping.")
    parser.add_argument("--root", default="qscaild-master/test_dir", help="Path to qscaild reference test_dir.")
    args = parser.parse_args()

    root = Path(args.root).resolve()
    if not root.exists():
        print(f"Reference root not found: {root}")
        return 1

    files = sorted(root.rglob("parameters"))
    if not files:
        print(f"No 'parameters' files found under: {root}")
        return 1

    print(f"Found {len(files)} parameter files under {root}")
    print("")
    for p in files:
        params = parse_parameters(p)
        mapped = map_to_macer_cli(params)
        rel = p.relative_to(root)
        print(f"[{rel}]")
        print("  qscaild params:")
        for key in ["T_K", "nconf", "nfits", "memory", "mixing", "tolerance", "pdiff", "grid", "imaginary_freq", "use_pressure", "pressure_diag", "use_smalldisp"]:
            if key in params:
                print(f"    {key} = {params[key]}")
        print("  macer qscaild suggestion:")
        if mapped:
            print("    macer phonopy qscaild -p POSCAR " + " ".join(mapped) + " --dry-run")
        else:
            print("    (no mapped options)")
        print("")

    print("Done.")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())




import numpy as np


def project_onto_wave_vector(trajectory, q_vector, project_on_atom=-1):

    number_of_primitive_atoms = trajectory.structure.get_number_of_primitive_atoms()
    velocity = trajectory.get_velocity_mass_average()
#    velocity = trajectory.velocity   # (use the velocity without mass average, just for testing)

    number_of_atoms = velocity.shape[1]
    number_of_dimensions = velocity.shape[2]
    supercell = trajectory.get_supercell_matrix()

    coordinates = trajectory.structure.get_positions(supercell)
    atom_type = trajectory.structure.get_atom_type_index(supercell=supercell)

    velocity_projected = np.zeros((velocity.shape[0], number_of_primitive_atoms, number_of_dimensions), dtype=complex)

    if q_vector.shape[0] != coordinates.shape[1]:
        print("Warning!! Q-vector and coordinates dimension do not match")
        exit()

    #Projection into wave vector
    for i in range(number_of_atoms):
        # Projection on atom
        if project_on_atom > -1:
            if atom_type[i] != project_on_atom:
                continue

        for k in range(number_of_dimensions):
            velocity_projected[:, atom_type[i], k] += velocity[:,i,k]*np.exp(-1j*np.dot(q_vector, coordinates[i,:]))

   #Normalize velocities (method 1)
  #  for i in range(velocity_projected.shape[1]):
  #      velocity_projected[:,i,:] /= atom_type.count(i)

   #Normalize velocities (method 2)
    number_of_primitive_cells = number_of_atoms/number_of_primitive_atoms
    velocity_projected /= np.sqrt(number_of_primitive_cells)
    return velocity_projected


def project_onto_phonon(vc, eigenvectors):

    number_of_cell_atoms = vc.shape[1]
    number_of_frequencies = eigenvectors.shape[0]

    #Projection in phonon coordinate
    velocity_projected=np.zeros((vc.shape[0], number_of_frequencies), dtype=complex)
    for k in range(number_of_frequencies):
        for i in range(number_of_cell_atoms):
            velocity_projected[:, k] += np.dot(vc[:, i, :], eigenvectors[k, i, :].conj())

    return velocity_projected


#Just for testing (slower implementation) [but equivalent]
def project_onto_phonon2(vc,eigenvectors):

    number_of_cell_atoms = vc.shape[1]
    number_of_frequencies = eigenvectors.shape[0]

    #Projection in phonon coordinate
    velocity_projected=np.zeros((vc.shape[0],number_of_frequencies),dtype=complex)

    for i in range (vc.shape[0]):
        for k in range(number_of_frequencies):
            velocity_projected[i,k] = np.trace(np.dot(vc[i,:,:], eigenvectors[k,:,:].T.conj()))
#            velocity_projected[i,k] = np.sum(np.linalg.eigvals(np.dot(vc[i,:,:],eigenvectors[k,:,:].T.conj())))
    return velocity_projected




#!/usr/bin/env python3
import argparse
import os
import sys
import yaml
import numpy as np
import matplotlib.pyplot as plt
from ase.io import read
from pathlib import Path

# Constants
KB_J = 1.380649e-23     # Boltzmann constant in J/K
E_CHARGE = 1.60217663e-19 # Elementary charge in C
ANGSTROM = 1e-10        # m
FEMTOSECOND = 1e-15     # s

def load_default_charges():
    """Load default oxidation states from pydefect database if available."""
    # 1. Try to find via pydefect package
    try:
        import pydefect
        pydefect_path = Path(pydefect.__file__).parent
        yaml_path = pydefect_path / "database" / "oxidation_state.yaml"
        if yaml_path.exists():
            with open(yaml_path, "r") as f:
                return yaml.safe_load(f)
    except (ImportError, Exception):
        pass

    # 2. Try relative to script (original logic)
    script_path = Path(__file__).resolve()
    project_root = script_path.parent.parent
    yaml_path = project_root / "pydefect" / "database" / "oxidation_state.yaml"
    
    if yaml_path.exists():
        try:
            with open(yaml_path, "r") as f:
                return yaml.safe_load(f)
        except Exception as e:
            print(f"Warning: Could not load oxidation states from {yaml_path}: {e}")
    return {}

def parse_user_charges(user_str, defaults):
    """Update defaults with user provided string 'El:Q,El2:Q2'."""
    charges = defaults.copy()
    if not user_str:
        return charges
    
    parts = user_str.split(',')
    for p in parts:
        p = p.strip()
        if ':' in p:
            el, q = p.split(':')
            try:
                charges[el.strip()] = float(q)
            except ValueError:
                print(f"Warning: Could not parse charge for '{p}'")
    return charges

def detect_interval(input_path):
    """Try to detect interval from XDATCAR configuration numbers."""
    if 'XDATCAR' not in str(input_path):
        return None
    try:
        configs = []
        with open(input_path, 'r') as f:
            for line in f:
                if "Direct configuration=" in line:
                    val = line.split('=')[1].strip().split()[0]
                    configs.append(int(val))
                if len(configs) >= 2:
                    break
        if len(configs) >= 2:
            return configs[1] - configs[0]
    except Exception as e:
        pass
    return None

def calculate_conductivity(args):
    input_path = Path(args.input)
    if not input_path.exists():
        print(f"Error: Input file {args.input} not found.")
        return

    # Auto-detect interval if not provided (or set to 1)
    if args.interval == 1:
        detected = detect_interval(input_path)
        if detected:
            args.interval = detected
            print(f"Auto-detected interval: {args.interval}")

    # Setup output path (in the same directory as input)
    out_prefix = input_path.parent / args.out
    log_file = f"{out_prefix}.log"
    
    def log_print(msg):
        print(msg)
        with open(log_file, "a") as f:
            f.write(msg + "\n")
            
    # Clear log file if exists
    with open(log_file, "w") as f:
        f.write("--- MD Conductivity Analysis ---\n")
        f.write(f"Input: {args.input}\n")
        f.write(f"Temp: {args.temp} K\n")
        f.write(f"Time step: {args.dt} fs\n")
        f.write(f"Interval: {args.interval}\n")
        if args.charge_msd:
            f.write("Mode: Charge MSD (correlated) enabled\n")
        f.write("-" * 40 + "\n")

    log_print(f"Reading trajectory: {args.input}")
    
    try:
        # Auto-detect format: use 'vasp-xdatcar' only if 'XDATCAR' is in the name
        fmt = 'vasp-xdatcar' if 'XDATCAR' in str(args.input) else None
        from ase.io import iread, read
        import io
        traj = []
        
        try:
            for atoms in iread(str(args.input), format=fmt):
                traj.append(atoms)
        except Exception as e:
            if fmt == 'vasp-xdatcar':
                log_print(f"Standard XDATCAR reading failed: {e}")
                log_print("Attempting robust reading mode for concatenated XDATCAR...")
                with open(str(args.input), 'r') as f:
                    content = f.read()
                
                if not content.strip():
                    raise ValueError("File is empty")
                    
                first_line = content.splitlines()[0]
                pieces = content.split(first_line + '\n')
                traj = []
                for p in pieces:
                    if not p.strip():
                        continue
                    try:
                        frames = read(io.StringIO(first_line + '\n' + p), index=':', format='vasp-xdatcar')
                        traj.extend(frames)
                    except:
                        continue
                if not traj:
                    raise e
                log_print(f"Successfully read {len(traj)} frames in robust mode.")
            else:
                raise e
    except Exception as e:
        log_print(f"Error reading trajectory: {e}")
        return

    if not traj:
        log_print("Error: Trajectory is empty.")
        return

    n_steps = len(traj)
    n_atoms = len(traj[0])
    species = sorted(list(set(traj[0].get_chemical_symbols())))
    
    vol_A3 = traj[0].get_volume()
    vol_m3 = vol_A3 * (ANGSTROM ** 3)
    
    log_print(f"Steps: {n_steps}, Atoms: {n_atoms}, Volume: {vol_A3:.2f} A^3")
    log_print(f"Species found: {species}")
    
    # Charges
    defaults = load_default_charges()
    charges = parse_user_charges(args.charges, defaults)
    
    log_print("Using oxidation states:")
    for sp in species:
        q = charges.get(sp, None)
        if q is None:
            q = 0.0
            log_print(f"  {sp}: {q} (Warning: not found in defaults, using 0.0)")
        else:
            log_print(f"  {sp}: {q}")
        charges[sp] = q
        
    # Unwrapping
    log_print("Unwrapping coordinates...")
    pos_list = [atoms.get_scaled_positions() for atoms in traj]
    scaled_pos = np.stack(pos_list)
    d_scaled = scaled_pos[1:] - scaled_pos[:-1]
    # Apply Minimum Image Convention
    d_scaled -= np.round(d_scaled)
    
    cum_disp_scaled = np.zeros_like(scaled_pos)
    cum_disp_scaled[1:] = np.cumsum(d_scaled, axis=0)
    
    # Handle Cell (Variable or Constant)
    cell_0 = traj[0].get_cell()
    cell_last = traj[-1].get_cell()
    if np.allclose(cell_0, cell_last, atol=1e-3):
        cell = cell_0
        unwrapped_real = np.dot(cum_disp_scaled + scaled_pos[0], cell)
    else:
        log_print("Variable cell detected. Transforming per step (approximation)...")
        unwrapped_real = np.zeros_like(scaled_pos)
        # Project accumulated scaled positions onto the instantaneous cell
        for t in range(n_steps):
            unwrapped_real[t] = (cum_disp_scaled[t] + scaled_pos[0]) @ traj[t].get_cell()

    # Effective time step
    dt_effective = args.dt * args.interval
    time_ps = np.arange(n_steps) * dt_effective * 1e-3

    # MSD & Sigma
    results = {}
    
    # --------------------------------------------------
    # Charge MSD preparation (Einstein form)
    # --------------------------------------------------
    if args.charge_msd:
        log_print("\nComputing charge MSD (Einstein form)...")

        symbols = traj[0].get_chemical_symbols()
        charges_SI = np.zeros(n_atoms)

        for i, sp in enumerate(symbols):
            charges_SI[i] = charges.get(sp, 0.0) * E_CHARGE

        # q_i * (r_i(t) - r_i(0))
        disp_charge = np.zeros((n_steps, 3))
        for t in range(n_steps):
            dr = unwrapped_real[t] - unwrapped_real[0]
            disp_charge[t] = np.sum(dr * charges_SI[:, None], axis=0)

        # | Σ q_i (r_i(t) - r_i(0)) |^2
        charge_msd = np.sum(disp_charge**2, axis=1)

    plt.figure(figsize=(8, 6))
    
    for sp in species:
        indices = [i for i, s in enumerate(traj[0].get_chemical_symbols()) if s == sp]
        if not indices:
            continue
            
        n_ions = len(indices)
        pos_sp = unwrapped_real[:, indices, :]
        
        # MSD(t) = < |r(t) - r(0)|^2 >
        disp_from_0 = pos_sp - pos_sp[0]
        sq_dist = np.sum(disp_from_0**2, axis=2)
        msd = np.mean(sq_dist, axis=1)
        
        # Fit (10% to 90% range to avoid initial ballistic and poor statistics at end)
        start_idx = int(n_steps * 0.1)
        end_idx = int(n_steps * 0.9)
        if start_idx >= end_idx:
            start_idx, end_idx = 0, n_steps
            
        fit_time = time_ps[start_idx:end_idx]
        fit_msd = msd[start_idx:end_idx]
        
        slope = 0
        intercept = 0
        if len(fit_time) > 1:
            slope, intercept = np.polyfit(fit_time, fit_msd, 1)
            
        # D in cm^2/s
        # MSD = 6Dt => D = slope / 6
        D_A2_ps = slope / 6.0
        D_SI = D_A2_ps * 1e-8 # m^2/s (1e-20 / 1e-12)
        D_cm2_s = D_SI * 1e4
        
        # Sigma in S/m
        # sigma = (N/V) * (q^2 * D) / (kT)
        q = charges.get(sp, 0.0) * E_CHARGE
        n_density = n_ions / vol_m3
        
        if q == 0:
            sigma = 0.0
        else:
            sigma = (n_density * (q**2) * D_SI) / (KB_J * args.temp)
            
        results[sp] = {'D': D_cm2_s, 'sigma': sigma, 'msd': msd}
        
        log_print(f"\n--- Species: {sp} ---")
        log_print(f"  Count: {n_ions}")
        log_print(f"  Diff. Coeff (D): {D_cm2_s:.3e} cm^2/s")
        log_print(f"  Conductivity (sigma_NE): {sigma:.3e} S/m ({sigma*10:.2f} mS/cm)")
        
        # Plotting
        plt.plot(time_ps, msd, label=f"{sp} ($D$={D_cm2_s:.1e})")
        if len(fit_time) > 1:
            plt.plot(fit_time, slope * fit_time + intercept, '--', alpha=0.5, color='gray')

    # --------------------------------------------------
    # Ionic conductivity from charge MSD (correlated)
    # --------------------------------------------------
    sigma_charge = None

    if args.charge_msd:
        start_idx = int(n_steps * 0.1)
        end_idx = int(n_steps * 0.9)
        if start_idx >= end_idx:
            start_idx, end_idx = 0, n_steps

        fit_time = time_ps[start_idx:end_idx]
        fit_charge_msd = charge_msd[start_idx:end_idx]

        if len(fit_time) > 1:
            slope_charge, _ = np.polyfit(fit_time, fit_charge_msd, 1)
        else:
            slope_charge = 0.0

        # (C·Å)^2 → (C·m)^2
        slope_charge_SI = slope_charge * (ANGSTROM ** 2)

        sigma_charge = slope_charge_SI / (6 * vol_m3 * KB_J * args.temp)

        log_print("\n--- Charge MSD conductivity (correlated) ---")
        log_print(f"  sigma_charge = {sigma_charge:.3e} S/m ({sigma_charge*10:.2f} mS/cm)")

    total_sigma = sum(r['sigma'] for r in results.values())

    log_print("\n===== Conductivity Summary =====")
    log_print(f"Total sigma_NE: {total_sigma:.3e} S/m ({total_sigma*10:.2f} mS/cm)")

    if sigma_charge is not None:
        haven = sigma_charge / total_sigma if total_sigma > 0 else 0.0
        log_print(f"Sigma from charge MSD: {sigma_charge:.3e} S/m ({sigma_charge*10:.2f} mS/cm)")
        log_print(f"Haven ratio (sigma_charge / sigma_NE): {haven:.3f}")
    
    plt.xlabel("Time (ps)")
    plt.ylabel(r"MSD ($\AA^2$)")
    plt.title(rf"MSD @ {args.temp}K (Total $\sigma$ = {total_sigma*10:.2f} mS/cm)")
    plt.legend()
    plt.grid(True, alpha=0.3)
    plt.tight_layout()
    
    plot_file = f"{out_prefix}.pdf"
    plt.savefig(plot_file)
    log_print(f"Plot saved to: {plot_file}")

    if args.charge_msd:
        plt.figure(figsize=(8, 6))
        plt.plot(time_ps, charge_msd)
        plt.xlabel("Time (ps)")
        plt.ylabel(r"Charge MSD $(C\cdot \AA)^2$")
        plt.title("Charge MSD (Einstein conductivity)")
        plt.grid(True, alpha=0.3)
        plt.tight_layout()
        plt.savefig(f"{out_prefix}_charge_msd.pdf")
        log_print(f"Charge MSD plot saved to: {out_prefix}_charge_msd.pdf")

    log_print(f"Log saved to: {log_file}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Calculate MSD and Ionic Conductivity from trajectory (XDATCAR or md.traj).",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("input", nargs="?", default="XDATCAR", help="Path to trajectory file (XDATCAR or md.traj)")
    parser.add_argument("--dt", type=float, default=2.0, help="MD Time step in fs")
    parser.add_argument("--temp", type=float, default=600.0, help="Temperature in K")
    parser.add_argument("--interval", type=int, default=1, help="Sampling interval (stride) of XDATCAR vs MD steps")
    parser.add_argument("--out", default="md_results", help="Output prefix for log and plot files")
    parser.add_argument("--charges", type=str, default="", help="Override oxidation states, e.g., 'Li:1,P:5,S:-2'")
    parser.add_argument(
        "--charge_msd",
        action="store_true",
        help="Also compute ionic conductivity from charge MSD (Einstein form)."
    )
    
    args = parser.parse_args()
    
    calculate_conductivity(args)


from argparse import ArgumentParser

import torch

from macer.externals.mace_bundled.tools.scripts_utils import remove_pt_head


def main():
    parser = ArgumentParser()
    grp = parser.add_mutually_exclusive_group()
    grp.add_argument(
        "--head_name",
        "-n",
        help="name of the head to extract",
        default=None,
    )
    grp.add_argument(
        "--list_heads",
        "-l",
        action="store_true",
        help="list names of the heads",
    )
    parser.add_argument(
        "--target_device",
        "-d",
        help="target device, defaults to model's current device",
    )
    parser.add_argument(
        "--output_file",
        "-o",
        help="name for output model, defaults to model.head_name, followed by .target_device if specified",
    )
    parser.add_argument("model_file", help="input model file path")
    args = parser.parse_args()

    model = torch.load(args.model_file, map_location=args.target_device)
    torch.set_default_dtype(next(model.parameters()).dtype)

    if args.list_heads:
        print("Available heads:")
        print("\n".join(["  " + h for h in model.heads]))
    else:

        if args.output_file is None:
            args.output_file = (
                args.model_file
                + "."
                + args.head_name
                + ("." + args.target_device if (args.target_device is not None) else "")
            )

        model_single = remove_pt_head(model, args.head_name)
        if args.target_device is not None:
            target_device = str(next(model.parameters()).device)
            model_single.to(target_device)
        torch.save(model_single, args.output_file)


if __name__ == "__main__":
    main()



###########################################################################################
# Radial basis and cutoff
# Authors: Ilyes Batatia, Gregor Simm
# This program is distributed under the MIT License (see MIT.md)
###########################################################################################

import logging

import ase
import numpy as np
import torch
from macer.externals.e3nn_legacy.util.jit import compile_mode

from macer.externals.mace_bundled.tools.scatter import scatter_sum


@compile_mode("script")
class BesselBasis(torch.nn.Module):
    """
    Equation (7)
    """

    def __init__(self, r_max: float, num_basis=8, trainable=False):
        super().__init__()

        bessel_weights = (
            np.pi
            / r_max
            * torch.linspace(
                start=1.0,
                end=num_basis,
                steps=num_basis,
                dtype=torch.get_default_dtype(),
            )
        )
        if trainable:
            self.bessel_weights = torch.nn.Parameter(bessel_weights)
        else:
            self.register_buffer("bessel_weights", bessel_weights)

        self.register_buffer(
            "r_max", torch.tensor(r_max, dtype=torch.get_default_dtype())
        )
        self.register_buffer(
            "prefactor",
            torch.tensor(np.sqrt(2.0 / r_max), dtype=torch.get_default_dtype()),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # [..., 1]
        numerator = torch.sin(self.bessel_weights * x)  # [..., num_basis]
        return self.prefactor * (numerator / x)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(r_max={self.r_max}, num_basis={len(self.bessel_weights)}, "
            f"trainable={self.bessel_weights.requires_grad})"
        )


@compile_mode("script")
class ChebychevBasis(torch.nn.Module):
    """
    Equation (7)
    """

    def __init__(self, r_max: float, num_basis=8):
        super().__init__()
        self.register_buffer(
            "n",
            torch.arange(1, num_basis + 1, dtype=torch.get_default_dtype()).unsqueeze(
                0
            ),
        )
        self.num_basis = num_basis
        self.r_max = r_max

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # [..., 1]
        x = x.repeat(1, self.num_basis)
        n = self.n.repeat(len(x), 1)
        return torch.special.chebyshev_polynomial_t(x, n)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(r_max={self.r_max}, num_basis={self.num_basis},"
        )


@compile_mode("script")
class GaussianBasis(torch.nn.Module):
    """
    Gaussian basis functions
    """

    def __init__(self, r_max: float, num_basis=128, trainable=False):
        super().__init__()
        gaussian_weights = torch.linspace(
            start=0.0, end=r_max, steps=num_basis, dtype=torch.get_default_dtype()
        )
        if trainable:
            self.gaussian_weights = torch.nn.Parameter(
                gaussian_weights, requires_grad=True
            )
        else:
            self.register_buffer("gaussian_weights", gaussian_weights)
        self.coeff = -0.5 / (r_max / (num_basis - 1)) ** 2

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # [..., 1]
        x = x - self.gaussian_weights
        return torch.exp(self.coeff * torch.pow(x, 2))


@compile_mode("script")
class PolynomialCutoff(torch.nn.Module):
    """Polynomial cutoff function that goes from 1 to 0 as x goes from 0 to r_max.
    Equation (8) -- TODO: from where?
    """

    p: torch.Tensor
    r_max: torch.Tensor

    def __init__(self, r_max: float, p=6):
        super().__init__()
        self.register_buffer("p", torch.tensor(p, dtype=torch.int))
        self.register_buffer(
            "r_max", torch.tensor(r_max, dtype=torch.get_default_dtype())
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.calculate_envelope(x, self.r_max, self.p.to(torch.int))

    @staticmethod
    def calculate_envelope(
        x: torch.Tensor, r_max: torch.Tensor, p: torch.Tensor
    ) -> torch.Tensor:
        r_over_r_max = x / r_max
        envelope = (
            1.0
            - ((p + 1.0) * (p + 2.0) / 2.0) * torch.pow(r_over_r_max, p)
            + p * (p + 2.0) * torch.pow(r_over_r_max, p + 1)
            - (p * (p + 1.0) / 2) * torch.pow(r_over_r_max, p + 2)
        )
        return envelope * (x < r_max)

    def __repr__(self):
        return f"{self.__class__.__name__}(p={self.p}, r_max={self.r_max})"


@compile_mode("script")
class ZBLBasis(torch.nn.Module):
    """Implementation of the Ziegler-Biersack-Littmark (ZBL) potential
    with a polynomial cutoff envelope.
    """

    p: torch.Tensor

    def __init__(self, p=6, trainable=False, **kwargs):
        super().__init__()
        if "r_max" in kwargs:
            logging.warning(
                "r_max is deprecated. r_max is determined from the covalent radii."
            )

        # Pre-calculate the p coefficients for the ZBL potential
        self.register_buffer(
            "c",
            torch.tensor(
                [0.1818, 0.5099, 0.2802, 0.02817], dtype=torch.get_default_dtype()
            ),
        )
        self.register_buffer("p", torch.tensor(p, dtype=torch.int))
        self.register_buffer(
            "covalent_radii",
            torch.tensor(
                ase.data.covalent_radii,
                dtype=torch.get_default_dtype(),
            ),
        )
        if trainable:
            self.a_exp = torch.nn.Parameter(torch.tensor(0.300, requires_grad=True))
            self.a_prefactor = torch.nn.Parameter(
                torch.tensor(0.4543, requires_grad=True)
            )
        else:
            self.register_buffer("a_exp", torch.tensor(0.300))
            self.register_buffer("a_prefactor", torch.tensor(0.4543))

    def forward(
        self,
        x: torch.Tensor,
        node_attrs: torch.Tensor,
        edge_index: torch.Tensor,
        atomic_numbers: torch.Tensor,
    ) -> torch.Tensor:
        sender = edge_index[0]
        receiver = edge_index[1]
        node_atomic_numbers = atomic_numbers[torch.argmax(node_attrs, dim=1)].unsqueeze(
            -1
        )
        Z_u = node_atomic_numbers[sender].to(torch.int64)
        Z_v = node_atomic_numbers[receiver].to(torch.int64)
        a = (
            self.a_prefactor
            * 0.529
            / (torch.pow(Z_u, self.a_exp) + torch.pow(Z_v, self.a_exp))
        )
        r_over_a = x / a
        phi = (
            self.c[0] * torch.exp(-3.2 * r_over_a)
            + self.c[1] * torch.exp(-0.9423 * r_over_a)
            + self.c[2] * torch.exp(-0.4028 * r_over_a)
            + self.c[3] * torch.exp(-0.2016 * r_over_a)
        )
        v_edges = (14.3996 * Z_u * Z_v) / x * phi
        r_max = self.covalent_radii[Z_u] + self.covalent_radii[Z_v]
        envelope = PolynomialCutoff.calculate_envelope(x, r_max, self.p)
        v_edges = 0.5 * v_edges * envelope
        V_ZBL = scatter_sum(v_edges, receiver, dim=0, dim_size=node_attrs.size(0))
        return V_ZBL.squeeze(-1)

    def __repr__(self):
        return f"{self.__class__.__name__}(c={self.c})"


@compile_mode("script")
class AgnesiTransform(torch.nn.Module):
    """Agnesi transform - see section on Radial transformations in
    ACEpotentials.jl, JCP 2023 (https://doi.org/10.1063/5.0158783).
    """

    def __init__(
        self,
        q: float = 0.9183,
        p: float = 4.5791,
        a: float = 1.0805,
        trainable=False,
    ):
        super().__init__()
        self.register_buffer("q", torch.tensor(q, dtype=torch.get_default_dtype()))
        self.register_buffer("p", torch.tensor(p, dtype=torch.get_default_dtype()))
        self.register_buffer("a", torch.tensor(a, dtype=torch.get_default_dtype()))
        self.register_buffer(
            "covalent_radii",
            torch.tensor(
                ase.data.covalent_radii,
                dtype=torch.get_default_dtype(),
            ),
        )
        if trainable:
            self.a = torch.nn.Parameter(torch.tensor(1.0805, requires_grad=True))
            self.q = torch.nn.Parameter(torch.tensor(0.9183, requires_grad=True))
            self.p = torch.nn.Parameter(torch.tensor(4.5791, requires_grad=True))

    def forward(
        self,
        x: torch.Tensor,
        node_attrs: torch.Tensor,
        edge_index: torch.Tensor,
        atomic_numbers: torch.Tensor,
    ) -> torch.Tensor:
        sender = edge_index[0]
        receiver = edge_index[1]
        node_atomic_numbers = atomic_numbers[torch.argmax(node_attrs, dim=1)].unsqueeze(
            -1
        )
        Z_u = node_atomic_numbers[sender].to(torch.int64)
        Z_v = node_atomic_numbers[receiver].to(torch.int64)
        r_0: torch.Tensor = 0.5 * (self.covalent_radii[Z_u] + self.covalent_radii[Z_v])
        r_over_r_0 = x / r_0
        return (
            1
            + (
                self.a
                * torch.pow(r_over_r_0, self.q)
                / (1 + torch.pow(r_over_r_0, self.q - self.p))
            )
        ).reciprocal_()

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(a={self.a:.4f}, q={self.q:.4f}, p={self.p:.4f})"
        )


@compile_mode("script")
class SoftTransform(torch.nn.Module):
    """
    Tanh-based smooth transformation:
        T(x) = p1 + (x - p1)*0.5*[1 + tanh(alpha*(x - m))],
    which smoothly transitions from ~p1 for x << p1 to ~x for x >> r0.
    """

    def __init__(self, alpha: float = 4.0, trainable=False):
        """
        Args:
            p1 (float): Lower "clamp" point.
            alpha (float): Steepness; if None, defaults to ~6/(r0-p1).
            trainable (bool): Whether to make parameters trainable.
        """
        super().__init__()
        # Initialize parameters
        self.register_buffer(
            "alpha", torch.tensor(alpha, dtype=torch.get_default_dtype())
        )
        if trainable:
            self.alpha = torch.nn.Parameter(self.alpha.clone())
        self.register_buffer(
            "covalent_radii",
            torch.tensor(
                ase.data.covalent_radii,
                dtype=torch.get_default_dtype(),
            ),
        )

    def compute_r_0(
        self,
        node_attrs: torch.Tensor,
        edge_index: torch.Tensor,
        atomic_numbers: torch.Tensor,
    ) -> torch.Tensor:
        """
        Compute r_0 based on atomic information.

        Args:
            node_attrs (torch.Tensor): Node attributes (one-hot encoding of atomic numbers).
            edge_index (torch.Tensor): Edge index indicating connections.
            atomic_numbers (torch.Tensor): Atomic numbers.

        Returns:
            torch.Tensor: r_0 values for each edge.
        """
        sender = edge_index[0]
        receiver = edge_index[1]
        node_atomic_numbers = atomic_numbers[torch.argmax(node_attrs, dim=1)].unsqueeze(
            -1
        )
        Z_u = node_atomic_numbers[sender].to(torch.int64)
        Z_v = node_atomic_numbers[receiver].to(torch.int64)
        r_0: torch.Tensor = self.covalent_radii[Z_u] + self.covalent_radii[Z_v]
        return r_0

    def forward(
        self,
        x: torch.Tensor,
        node_attrs: torch.Tensor,
        edge_index: torch.Tensor,
        atomic_numbers: torch.Tensor,
    ) -> torch.Tensor:

        r_0 = self.compute_r_0(node_attrs, edge_index, atomic_numbers)
        p_0 = (3 / 4) * r_0
        p_1 = (4 / 3) * r_0
        m = 0.5 * (p_0 + p_1)
        alpha = self.alpha / (p_1 - p_0)
        s_x = 0.5 * (1.0 + torch.tanh(alpha * (x - m)))
        return p_0 + (x - p_0) * s_x

    def __repr__(self):
        return f"{self.__class__.__name__}(alpha={self.alpha.item():.4f})"


class RadialMLP(torch.nn.Module):
    """
    Construct a radial MLP (Linear → LayerNorm → SiLU) stack
    given a list of channel sizes, following ESEN / FairChem.
    """

    def __init__(self, channels_list) -> None:
        super().__init__()

        modules = []
        in_channels = channels_list[0]

        for idx, out_channels in enumerate(channels_list[1:], start=1):
            modules.append(torch.nn.Linear(in_channels, out_channels, bias=True))
            in_channels = out_channels
            if idx < len(channels_list) - 1:
                modules.append(torch.nn.LayerNorm(out_channels))
                modules.append(torch.nn.SiLU())

        self.net = torch.nn.Sequential(*modules)
        self.hs = channels_list

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        return self.net(inputs)



#!/usr/bin/env python3
import sys
import os
import argparse
from argparse import Namespace
from pathlib import Path
import yaml
import warnings

# Add current directory to sys.path
sys.path.insert(0, str(Path.cwd()))

from pymatgen.core import IStructure
from monty.serialization import loadfn

# pydefect imports
from pydefect.cli.main_functions import (
    make_supercell,
    make_defect_energy_infos_main_func,
    make_defect_energy_summary_main_func
)
from pydefect.cli.vasp.main_vasp_functions import make_defect_entries, make_calc_results
from pydefect.input_maker.defect_set_maker import DefectSetMaker
from pydefect.chem_pot_diag.chem_pot_diag import StandardEnergies, TargetVertices
from pydefect.analyzer.unitcell import Unitcell
from pydefect.analyzer.band_edge_states import PerfectBandEdgeState, EdgeInfo, OrbitalInfo
from pydefect.analyzer.defect_energy import DefectEnergyInfo
from pydefect.analyzer.make_defect_structure_info import MakeDefectStructureInfo
from pydefect.defaults import defaults
from pymatgen.core import Structure

# macer imports
from macer.calculator.factory import get_available_ffs, get_calculator
from macer.relaxation.optimizer import relax_structure
from macer.utils.logger import Logger
from macer.io.writers import write_pydefect_dummy_files
from macer.defaults import DEFAULT_MODELS, _macer_root, DEFAULT_DEVICE

# warnings
warnings.filterwarnings("ignore")

def write_summary_at_vertices(summary_path: Path):
    """
    Calculates and prints defect formation energies for each chemical potential
    vertex from a defect_energy_summary.json file.
    """
    if not summary_path.exists():
        print(f"Error: {summary_path} not found.")
        return

    try:
        # loadfn will deserialize into DefectEnergySummary object
        summary = loadfn(summary_path)
    except Exception as e:
        print(f"Error loading {summary_path}: {e}")
        return

    rel_chem_pots = summary.rel_chem_pots
    defect_energies_summary = summary.defect_energies
    output_filename = "defect_energy_summary-at-verticies.txt"

    with open(output_filename, "w") as f:
        f.write(f"title: {summary.title}\n\n")

        for vertex_label, chem_pots in rel_chem_pots.items():
            f.write(f"--- Vertex: {vertex_label} ---\n")
            chem_pot_str = ", ".join([f"{elem}: {pot:.4f}" for elem, pot in chem_pots.items()])
            f.write(f"Relative chemical potentials: {chem_pot_str}\n")
            f.write("-" * 40 + "\n")
            f.write(f"{'Defect Name':<20} {'Formation Energy (eV)'}\n")
            f.write(f"{'-'*20} {'-'*25}\n")

            formation_energies = {}
            for defect_name, defect_group in defect_energies_summary.items():
                charge_0_defect_energy = None
                try:
                    charge_0_index = defect_group.charges.index(0)
                    charge_0_defect_energy = defect_group.defect_energies[charge_0_index]
                except (ValueError, AttributeError):
                    print(f"Warning: Defect {defect_name} does not have a charge 0 state in the summary. Skipping.")
                    continue

                if not charge_0_defect_energy:
                    continue

                atom_io = defect_group.atom_io
                base_energy = charge_0_defect_energy.formation_energy
                
                reservoir_contribution = 0.0
                for element, num_diff in atom_io.items():
                    if element in chem_pots:
                        reservoir_contribution -= num_diff * chem_pots[element]

                formation_energy = base_energy + reservoir_contribution
                formation_energies[defect_name] = formation_energy
            
            sorted_defects = sorted(formation_energies.items(), key=lambda item: item[1])
            for name, energy in sorted_defects:
                f.write(f"{name:<20} {energy:>25.4f}\n")
            
            f.write("\n")

    print(f"Output written to {output_filename}")


def main():
    parser = argparse.ArgumentParser(description="Auto Defect Workflow with Macer")
    parser.add_argument("-p", "--poscar", type=str, required=True, help="Unitcell POSCAR file")
    parser.add_argument("-s", "--std_energies", type=str, required=True, help="Path to standard_energies.yaml")
    parser.add_argument("-t", "--target_vertices", type=str, required=True, help="Path to target_vertices.yaml")
    args = parser.parse_args()

    poscar_path = Path(args.poscar)
    if not poscar_path.exists():
        print(f"Error: {poscar_path} not found.")
        sys.exit(1)
    
    std_energies_path = Path(args.std_energies)
    if not std_energies_path.exists():
        print(f"Error: {std_energies_path} not found.")
        sys.exit(1)

    target_vertices_path = Path(args.target_vertices)
    if not target_vertices_path.exists():
        print(f"Error: {target_vertices_path} not found.")
        sys.exit(1)

    print("--- Step 1: Make Supercell ---")
    unitcell = IStructure.from_file(str(poscar_path))
    args_sc = Namespace(unitcell=unitcell, matrix=None, min_num_atoms=50, max_num_atoms=300, 
                        analyze_symmetry=True, sites_yaml_filename=None)
    make_supercell(args_sc)
    
    if not Path("supercell_info.json").exists():
        print("Error: supercell_info.json not created.")
        sys.exit(1)

    print("\n--- Step 2: Make Defect Set (Charge 0 only) ---")
    supercell_info = loadfn("supercell_info.json")
    maker = DefectSetMaker(supercell_info)
    
    defect_dict = {defect.name: [0] for defect in maker.defect_set}
    
    with open("defect_in.yaml", "w") as f:
        yaml.dump(defect_dict, f)
    print("defect_in.yaml created with charge states [0].")

    print("\n--- Step 3: Make Defect Entries ---")
    make_defect_entries(Namespace())
    
    dirs_to_calc = ["perfect"] + [f"{name}_0" for name in defect_dict if Path(f"{name}_0").exists()]
    print(f"Directories to calculate: {dirs_to_calc}")

    print("\n--- Step 4: Run Macer Relax ---")
    
    available_ffs = get_available_ffs()
    if not available_ffs:
        print("Error: No MLFF packages installed.")
        return
    
    ff = available_ffs[0]
    device = DEFAULT_DEVICE
    
    default_model_name = DEFAULT_MODELS.get(ff, "")
    model_path = os.path.join(_macer_root, "mlff-model", default_model_name) if default_model_name and ff not in {"fairchem", "orb", "chgnet", "m3gnet"} else default_model_name

    print(f"Initializing {ff.upper()} calculator on {device}...")
    calc_kwargs = {"device": device}
    if ff == "mace":
        calc_kwargs["model_paths"] = [model_path]
    else:
        calc_kwargs["model_path"] = model_path

    try:
        calculator = get_calculator(ff_name=ff, **calc_kwargs)
    except Exception as e:
        print(f"Failed to initialize calculator: {e}")
        return

    cwd = Path.cwd()
    successful_dirs = []
    
    for d_name in dirs_to_calc:
        d = cwd / d_name
        if not d.exists():
            print(f"Skipping {d_name}, directory not found.")
            continue
            
        print(f"Relaxing in {d_name}...")
        os.chdir(d)
        try:
            write_pydefect_dummy_files(".")
            relax_structure(input_file="POSCAR", fmax=0.03, isif=2, device=device, calculator=calculator, ff=ff,
                            outcar_name="OUTCAR", contcar_name="CONTCAR", xml_name="vasprun.xml", make_pdf=False, write_json=True)
            if Path("vasprun.xml").exists() and Path("OUTCAR").exists():
                successful_dirs.append(d)

                # Generate defect_structure_info.json directly using CONTCAR
                try:
                    if Path("CONTCAR").exists() and Path("defect_entry.json").exists():
                        final_structure = Structure.from_file("CONTCAR")
                        defect_entry = loadfn("defect_entry.json")
                        
                        dsi_maker = MakeDefectStructureInfo(
                            perfect=supercell_info.structure,
                            initial=defect_entry.structure,
                            final=final_structure,
                            symprec=defaults.symmetry_length_tolerance,
                            dist_tol=defaults.dist_tol
                        )
                        dsi_maker.defect_structure_info.to_json_file("defect_structure_info.json")
                        print(f"Generated defect_structure_info.json for {d_name}")
                except Exception as e:
                    print(f"Warning: Failed to generate defect_structure_info.json for {d_name}: {e}")

        except Exception as e:
            print(f"Relaxation failed in {d_name}: {e}")
        finally:
            os.chdir(cwd)

    print("\n--- Step 4.5: Generate calc_results.json ---")
    if successful_dirs:
        make_calc_results(Namespace(dirs=[d for d in successful_dirs], verbose=False, check_calc_results=False))
    else:
        print("No successful calculations found.")
        return

    print("\n--- Step 5: Generate Defect Energy Info ---")
    
    dummy_unitcell_str = """system: dummy
vbm: 0.0
cbm: 5.0
ele_dielectric_const: [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
ion_dielectric_const: [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
"""
    Path("unitcell.yaml").write_text(dummy_unitcell_str)
    
    perfect_calc_results_path = Path("perfect/calc_results.json")
    if not perfect_calc_results_path.exists():
        print("Error: perfect/calc_results.json not found.")
        return

    perfect_calc_results = loadfn(str(perfect_calc_results_path))
    unitcell_obj = Unitcell.from_yaml("unitcell.yaml")
    std_energies_obj = StandardEnergies.from_yaml(str(std_energies_path))
    
    defect_dirs = [d for d in successful_dirs if d.name != "perfect"]
    
    args_dei = Namespace(dirs=defect_dirs, check_calc_results=False, perfect_calc_results=perfect_calc_results,
                         std_energies=std_energies_obj, unitcell=unitcell_obj, verbose=False)
    
    make_defect_energy_infos_main_func(args_dei)
    print("Done. Defect energy infos generated.")

    print("\n--- Step 6: Generate Defect Energy Summary ---")
    
    vbm_info = EdgeInfo(band_idx=0, kpt_coord=(0.0, 0.0, 0.0),
                        orbital_info=OrbitalInfo(energy=0.0, orbitals={}, occupation=1.0))
    cbm_info = EdgeInfo(band_idx=0, kpt_coord=(0.0, 0.0, 0.0),
                        orbital_info=OrbitalInfo(energy=5.0, orbitals={}, occupation=0.0))
    p_state = PerfectBandEdgeState(vbm_info=vbm_info, cbm_info=cbm_info)
    
    args_des = Namespace(
        dirs=defect_dirs,
        verbose=False,
        target_vertices_yaml=str(target_vertices_path),
        unitcell=unitcell_obj,
        p_state=p_state
    )
    
    make_defect_energy_summary_main_func(args_des)
    print("defect_energy_summary.json created.")

    print("\n--- Step 7: Write summary at each vertex ---")
    summary_path = Path("defect_energy_summary.json")
    if summary_path.exists():
        write_summary_at_vertices(summary_path)
    else:
        print("defect_energy_summary.json not found, skipping final summary generation.")


if __name__ == "__main__":
    main()



"""
Macer: Machine-learning accelerated Atomic Computational Environment for automated Research workflows
Copyright (c) 2025 The Macer Package Authors
Author: Soungmin Bae <soungminbae@gmail.com>
License: MIT
"""

import argparse
import sys
import glob
import copy
import os
from pathlib import Path

from macer.calculator.factory import get_available_ffs, ALL_SUPPORTED_FFS
from macer.defaults import DEFAULT_DEVICE, DEFAULT_FF
# from macer import __version__

MACER_LOGO = r"""
███╗   ███╗  █████╗   ██████╗ ███████╗ ██████╗
████╗ ████║ ██╔══██╗ ██╔════╝ ██╔════╝ ██╔══██╗
██╔████╔██║ ███████║ ██║      █████╗   ██████╔╝
██║╚██╔╝██║ ██╔══██║ ██║      ██╔══╝   ██╔══██╗
██║ ╚═╝ ██║ ██║  ██║ ╚██████╗ ███████╗ ██║  ██║
╚═╝     ╚═╝ ╚═╝  ╚═╝  ╚═════╝ ╚══════╝ ╚═╝  ╚═╝
macer: ML-accelerated Atomic Computational Environment for Research
"""

# Determine default force field based on installed extras
available_ffs = get_available_ffs()

def add_pydefect_parsers(subparsers):
    """Add pydefect sub-commands (cpd, defect, full) directly to the provided subparsers."""
    
    # CPD command
    cpd_epilog = """
Examples:
  # 1. Basic CPD generation using Materials Project ID (Recommended)
  macer pydefect cpd -m mp-3536

  # 2. CPD using formula (retrieves most stable from MP)
  macer pydefect cpd -f MgAl2O4

  # 3. CPD with a dopant element
  macer pydefect cpd -m mp-3536 -d Ca
"""
    cpd_parser = subparsers.add_parser(
        "cpd",
        description=MACER_LOGO + "\nRun Chemical Potential Diagram (CPD) workflow.",
        help="Generate CPD and target vertices.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=cpd_epilog
    )
    cpd_general = cpd_parser.add_argument_group('General & Input')
    cpd_general.add_argument("-f", "--formula", type=str, help="Formula to retrieve from Materials Project (e.g., MgAl2O4)")
    cpd_general.add_argument("-m", "--mpid", type=str, help="Materials Project ID (e.g., mp-3536)")
    cpd_general.add_argument("-d", "--doping", type=str, nargs='+', help="Dopant element(s) (e.g., Cl)")
    cpd_general.add_argument("-p", "--poscar", type=str, nargs='+', help="Input POSCAR file(s) (Optional)")
    _add_mlff_group(cpd_parser)
    cpd_calc = cpd_parser.add_argument_group('Calculation Settings')
    cpd_calc.add_argument("--fmax", type=float, default=0.03, help="Force convergence threshold (eV/Å). Default: 0.03")
    cpd_calc.add_argument("--energy-shift-target", type=float, default=0.0, help="Manually shift target energy (eV/atom). Default: 0.0")
    cpd_out = cpd_parser.add_argument_group('Output Settings')
    cpd_out.add_argument("--output-dir", type=str, help="Directory to save CPD output files.")
    cpd_parser.set_defaults(func=run_batch_workflow, workflow_name="cpd")

    # Defect command
    defect_epilog = """
Examples:
  # 1. Basic defect analysis with required inputs
  macer pydefect defect -p POSCAR --std_energies standard_energies.yaml --target_vertices target_vertices.yaml

  # 2. Defect analysis with a specific supercell matrix
  macer pydefect defect -p POSCAR -s standard_energies.yaml -t target_vertices.yaml --matrix 2 2 2

  # 3. Analysis with dopant and atom count limits
  macer pydefect defect -p POSCAR -s standard_energies.yaml -t target_vertices.yaml -d Mg --min_atoms 100 --max_atoms 400

  # 4. Tighter force convergence for defect relaxation
  macer pydefect defect -p POSCAR -s standard_energies.yaml -t target_vertices.yaml --fmax-defect 0.01

  # 5. Automatically find and add interstitial sites using Voronoi tessellation
  macer pydefect defect -p POSCAR -s standard_energies.yaml -t target_vertices.yaml --interstitial

  # 6. Disable symmetry analysis (if spglib fails or manual setup is preferred)
  macer pydefect defect -p POSCAR -s standard_energies.yaml -t target_vertices.yaml --no_symmetry_analysis
"""
    defect_parser = subparsers.add_parser(
        "defect",
        description=MACER_LOGO + "\nRun Defect Analysis workflow (Supercell generation, Relaxation, Analysis).",
        help="Calculate defect formation energies.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=defect_epilog
    )
    defect_general = defect_parser.add_argument_group('General & Input')
    defect_general.add_argument("-p", "--poscar", type=str, nargs='+', help="Input POSCAR file(s) (Perfect Unitcell)")
    defect_general.add_argument("-f", "--formula", type=str, help="Formula to retrieve from Materials Project (e.g., MgAl2O4)")
    defect_general.add_argument("-m", "--mpid", type=str, help="Materials Project ID (e.g., mp-3536)")
    defect_general.add_argument("-d", "--doping", type=str, nargs='+', help="Dopant element(s) (e.g., Cl)")
    defect_general.add_argument("-s", "--std_energies", type=str, help="Path to standard_energies.yaml")
    defect_general.add_argument("-t", "--target_vertices", type=str, help="Path to target_vertices.yaml")
    _add_mlff_group(defect_parser)
    defect_sc = defect_parser.add_argument_group('Supercell & Symmetry')
    defect_sc.add_argument("--matrix", nargs="+", type=int, help="Supercell matrix applied to the conventional cell.")
    defect_sc.add_argument("--dim", nargs="+", type=int, help="Supercell matrix applied directly to the input POSCAR (3 or 9 integers).")
    defect_sc.add_argument("--min_atoms", type=int, default=50, help="Minimum number of atoms (default: 50)")
    defect_sc.add_argument("--max_atoms", type=int, default=300, help="Maximum number of atoms (default: 300)")
    defect_sc.add_argument("--no_symmetry_analysis", dest="analyze_symmetry", action="store_false", help="Disable symmetry analysis.")
    defect_parser.set_defaults(analyze_symmetry=True)
    defect_sc.add_argument("--sites_yaml", type=str, dest="sites_yaml_filename", help="Path to sites.yaml file.")
    defect_sc.add_argument("--interstitial", action="store_true", help="Automatically find and add interstitial sites using Voronoi tessellation.")
    defect_relax = defect_parser.add_argument_group('Relaxation Settings')
    defect_relax.add_argument("--fmax", type=float, default=0.03, help="Force threshold (default: 0.03)")
    defect_relax.add_argument("--fmax-defect", type=float, help="Force threshold specifically for defect relaxation. Default: same as --fmax")
    defect_out = defect_parser.add_argument_group('Output Settings')
    defect_out.add_argument("--output-dir", type=str, help="Directory to save defect analysis output files.")
    defect_parser.set_defaults(func=run_batch_workflow, workflow_name="defect")

    # Full command
    full_epilog = """
Examples:
  # 1. Full defect workflow from scratch using MPID (Recommended)
  macer pydefect full -m mp-2605

  # 2. Full workflow using formula
  macer pydefect full -f MgO

  # 3. Full workflow with explicit input POSCAR and dopant
  macer pydefect full -p POSCAR_MgO -d Al

  # 4. Custom force thresholds for different stages
  macer pydefect full -m mp-2133 --fmax-cpd 0.02 --fmax-defect 0.05

  # 5. Full workflow with explicit supercell matrix
  macer pydefect full -m mp-149 --matrix 3 3 3

  # 6. Full workflow with automatic interstitial site finding
  macer pydefect full -m mp-2605 --interstitial
"""
    full_parser = subparsers.add_parser(
        "full",
        description=MACER_LOGO + "\nRun Full Defect Analysis workflow (CPD + Defect Analysis).",
        help="Run both CPD and Defect Analysis workflows.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=full_epilog
    )
    full_general = full_parser.add_argument_group('General & Input')
    full_general.add_argument("-f", "--formula", type=str, help="Formula (e.g., MgAl2O4)")
    full_general.add_argument("-m", "--mpid", type=str, help="Materials Project ID (e.g., mp-3536)")
    full_general.add_argument("-p", "--poscar", type=str, nargs='+', help="Input POSCAR file(s)")
    full_general.add_argument("-d", "--doping", type=str, nargs='+', help="Dopant element(s)")
    _add_mlff_group(full_parser)
    full_sc = full_parser.add_argument_group('Supercell & Symmetry')
    full_sc.add_argument("--matrix", nargs="+", type=int, help="Supercell matrix.")
    full_sc.add_argument("--dim", nargs="+", type=int, help="Supercell matrix applied directly to the input POSCAR (3 or 9 integers).")
    full_sc.add_argument("--min_atoms", type=int, default=50, help="Min atoms (default: 50)")
    full_sc.add_argument("--max_atoms", type=int, default=300, help="Max atoms (default: 300)")
    full_sc.add_argument("--no_symmetry_analysis", dest="analyze_symmetry", action="store_false", help="Disable symmetry analysis.")
    full_parser.set_defaults(analyze_symmetry=True)
    full_sc.add_argument("--sites_yaml", type=str, dest="sites_yaml_filename", help="Path to sites.yaml file.")
    full_sc.add_argument("--interstitial", action="store_true", help="Automatically find and add interstitial sites using Voronoi tessellation.")
    full_relax = full_parser.add_argument_group('Relaxation & Correction')
    full_relax.add_argument("--fmax", type=float, default=0.03, help="Global force threshold (default: 0.03)")
    full_relax.add_argument("--fmax-cpd", type=float, help="CPD force threshold.")
    full_relax.add_argument("--fmax-defect", type=float, help="Defect force threshold.")
    full_relax.add_argument("--energy-shift-target", type=float, default=0.0, help="Shift target energy. Default: 0.0")
    full_relax.add_argument("--without-cpd", action="store_true",
                            help="Skip CPD/target vertices; compute only elemental standard energies.")
    full_out = full_parser.add_argument_group('Output Settings')
    full_out.add_argument("--output-dir", type=str, help="Directory to save full defect analysis output files.")
    full_parser.set_defaults(func=run_batch_workflow, workflow_name="full")

    return {
        "cpd": cpd_parser,
        "defect": defect_parser,
        "full": full_parser
    }

def _add_mlff_group(parser):
    """Helper to add MLFF arguments to a parser group."""
    mlff_group = parser.add_argument_group('MLFF Model Settings')
    mlff_group.add_argument("--ff", type=str, default=DEFAULT_FF, choices=ALL_SUPPORTED_FFS, help=f"Force field to use. (default: {DEFAULT_FF})")
    mlff_group.add_argument("--model", type=str, default=None, help="Path to the MLFF model file.")
    mlff_group.add_argument("--device", type=str, default=DEFAULT_DEVICE, choices=["cpu", "mps", "cuda"], help="Compute device.")
    mlff_group.add_argument("--modal", type=str, default=None, help="Modal for SevenNet model.")
    mlff_group.add_argument("--n-jobs", type=int, default=1, help="Parallel workers for relaxations (CPU only, default: 1).")

def run_batch_workflow(args):
    workflow_name = args.workflow_name
    launch_cwd = Path.cwd()

    # Keep relative --model stable even if workflows chdir into work directories.
    if hasattr(args, "model") and args.model:
        model_path = Path(args.model)
        if not model_path.is_absolute():
            candidate = (launch_cwd / model_path).resolve()
            if candidate.exists():
                args.model = str(candidate)
    
    if workflow_name == "cpd":
        from macer.pydefect.cpd import run_cpd_workflow as workflow_func
    elif workflow_name == "defect":
        from macer.pydefect.defect import run_defect_workflow as workflow_func
    elif workflow_name == "full":
        from macer.pydefect.full import run_full_workflow as workflow_func
    else:
        raise ValueError(f"Unknown workflow: {workflow_name}")

    if workflow_name == "defect" and not (hasattr(args, 'poscar') and args.poscar):
        if getattr(args, "formula", None) or getattr(args, "mpid", None):
            from macer.utils.struct_tools import resolve_structure_inputs
            resolved = resolve_structure_inputs(
                poscar_paths=None,
                cif_paths=None,
                formula=getattr(args, "formula", None),
                mpid=getattr(args, "mpid", None), show_mp_progress=True)
            if resolved:
                args.poscar = resolved

    if hasattr(args, 'poscar') and args.poscar:
        input_patterns = args.poscar
        input_files = []
        for pat in input_patterns:
            if os.path.exists(pat): input_files.append(pat)
            else: input_files.extend(glob.glob(pat))
        input_files = sorted(list(set(input_files)))
        if not input_files:
            print(f"No input files found matching: {input_patterns}")
            sys.exit(1)
        original_cwd = os.getcwd()
        for f in input_files:
            os.chdir(original_cwd)
            print(f"\n{'='*60}\nProcessing Input File: {f}\n{'='*60}\n")
            single_args = copy.copy(args)
            single_args.poscar = f
            try: workflow_func(single_args)
            except Exception as e:
                print(f"Error processing {f}: {e}")
                import traceback
                traceback.print_exc()
                continue
    else: workflow_func(args)

def main():
    from macer import __version__
    parser = argparse.ArgumentParser(
        description=MACER_LOGO + f"\nmacer_pydefect (v{__version__}): Automated Point Defect Calculations with MLFFs",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("--version", "-v", action="version", version=f"macer_pydefect {__version__}")
    subparsers = parser.add_subparsers(dest="subcommand", help="Available commands")
    p_map = add_pydefect_parsers(subparsers)
    args = parser.parse_args()
    launch_cwd = Path.cwd()
    if hasattr(args, "model") and args.model:
        model_path = Path(args.model)
        if not model_path.is_absolute():
            candidate = (launch_cwd / model_path).resolve()
            if candidate.exists():
                args.model = str(candidate)

    if not args.subcommand:
        parser.print_help()
        return

    # Manual checks
    if args.subcommand == "cpd" and not (args.formula or args.mpid or args.poscar):
        p_map["cpd"].print_help(); return
    if args.subcommand == "defect" and not args.poscar:
        p_map["defect"].print_help(); return
    if args.subcommand == "full" and not (args.formula or args.mpid or args.poscar):
        p_map["full"].print_help(); return

    if hasattr(args, "func"): args.func(args)
    else: parser.print_help(); sys.exit(1)

if __name__ == "__main__":
    main()



"""Various base and utility definitions."""
# Copyright (C) 2015 Atsushi Togo
# This file is part of spglib.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import typing
import warnings
from collections.abc import Mapping, Sequence
from typing import Union

import numpy as np

import macer.externals.spglib_bundled as spglib

from . import _spglib
from ._compat.typing import TypeAlias
from ._compat.warnings import deprecated
from .error import SpglibError, _set_no_error

if typing.TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any

__all__ = [
    "Lattice",
    "Positions",
    "Numbers",
    "Magmoms",
    "Cell",
    "get_version",
    "spg_get_version",
    "spg_get_version_full",
    "spg_get_commit",
]

warnings.filterwarnings(
    "module", category=DeprecationWarning, message=r"dict interface.*"
)


Lattice: TypeAlias = Sequence[Sequence[float]]
"""Basis vectors in ``(3, 3)`` shape."""
Positions: TypeAlias = Sequence[Sequence[float]]
"""Point coordinates of atoms in ``(N, 3)`` shape."""
Numbers: TypeAlias = Sequence[int]
"""Integer numbers for identifying atoms in ``(N,)`` shape."""
Magmoms: TypeAlias = Union[Sequence[float], Sequence[Sequence[float]]]
"""Magnetic moments of atoms in either ``(N,)`` or ``(N, 3)`` shape."""

# Compatibility alias for some deprecated interfaces
Cell: TypeAlias = Union["spglib.spg.SpgCell", "spglib.msg.MsgCell"]
"""Either :py:type:`spglib.spg.SpgCell` or :py:type:`spglib.msg.MsgCell`."""


@dataclasses.dataclass(eq=True, frozen=True)
class DictInterface(Mapping[str, "Any"]):
    """Base class for dataclass with dict interface.

    .. versionadded:: 2.5.0
    .. deprecated:: 2.5.0
        Dict-like interface (``obj['field']``) are deprecated.
        Please use attribute interface instead (``obj.field``)
    """

    @deprecated("dict interface is deprecated. Use attribute interface instead")
    def __getitem__(self, key: str) -> Any:
        """Return the value of the key."""
        return dataclasses.asdict(self)[key]

    def __len__(self) -> int:
        """Return the number of fields."""
        return len(dataclasses.fields(self))

    def __iter__(self) -> Iterator[str]:
        """Return an iterator over the keys."""
        return iter(dataclasses.asdict(self))


@deprecated("Use __version__ or spg_get_version instead")
def get_version() -> tuple[int, int, int]:
    """Return version number of spglib with tuple of three numbers.

    .. versionadded:: 1.8.3
    .. deprecated:: 2.3.0
        Use :py:func:`spg_get_version` and ``spglib.__version__`` instead
    """
    _set_no_error()
    return _spglib.version_tuple()


def spg_get_version() -> str:
    """Get the X.Y.Z version of the detected spglib C library.

    .. versionadded:: 2.3.0

    :return: version string
    """
    _set_no_error()
    return _spglib.version_string()


def spg_get_version_full() -> str:
    """Get the full version of the detected spglib C library.

    .. versionadded:: 2.3.0

    :return: full version string
    """
    _set_no_error()
    return _spglib.version_full()


def spg_get_commit() -> str:
    """Get the commit of the detected spglib C library.

    .. versionadded:: 2.3.0

    :return: commit string
    """
    _set_no_error()
    return _spglib.commit()


def _expand_cell(
    cell: Cell,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray | None]:
    try:
        lattice = np.array(np.transpose(cell[0]), dtype="double", order="C")
        positions = np.array(cell[1], dtype="double", order="C")
        numbers = np.array(cell[2], dtype="intc")
        if len(cell) == 4:
            magmoms = np.array(cell[3], order="C", dtype="double")
        elif len(cell) == 3:
            magmoms = None
        else:
            raise TypeError("cell has to be a tuple of 3 or 4 elements.")

        # Sanity check
        if lattice.shape != (3, 3):
            raise TypeError("lattice has to be a (3, 3) array.")
        if not (positions.ndim == 2 and positions.shape[1] == 3):
            raise TypeError("positions has to be a (num_atoms, 3) array.")
        num_atoms = positions.shape[0]
        if numbers.ndim != 1:
            raise TypeError("numbers has to be a (num_atoms,) array.")
        if len(numbers) != num_atoms:
            raise TypeError(
                "numbers has to have the same number of atoms as positions."
            )
        if magmoms is not None:
            if len(magmoms) != num_atoms:
                raise TypeError(
                    "magmoms has to have the same number of atoms as positions."
                )
            if magmoms.ndim == 1:
                # collinear
                pass
            elif magmoms.ndim == 2:
                # non-collinear
                if magmoms.shape[1] != 3:
                    raise TypeError(
                        "non-collinear magmoms has to be a (num_atoms, 3) array."
                    )
            else:
                raise TypeError("magmoms has to be a 1D or 2D array.")
    except Exception as exc:
        # Note: these will eventually be moved to the C++ side
        # For now we just recast them to SpglibError
        raise SpglibError(f"Generic Spglib error:\n{exc}") from exc

    return (lattice, positions, numbers, magmoms)
